"""Correlation classes
======================

The correlation classes are able to compute correlations with
different methods.

.. autoclass:: CorrelBase
   :members:
   :private-members:

.. autoclass:: CorrelScipySignal
   :members:
   :private-members:

.. autoclass:: CorrelScipyNdimage
   :members:
   :private-members:

.. autoclass:: CorrelPythran
   :members:
   :private-members:

.. autoclass:: CorrelPyCuda
   :members:
   :private-members:

.. autoclass:: CorrelTheano
   :members:
   :private-members:

.. autoclass:: CorrelFFTNumpy
   :members:
   :private-members:

.. autoclass:: CorrelFFTW
   :members:
   :private-members:

.. autoclass:: CorrelCuFFT
   :members:
   :private-members:

.. autoclass:: CorrelSKCuFFT
   :members:
   :private-members:

"""

import numpy as np
from numpy.fft import fft2, ifft2
from scipy.ndimage import correlate
from scipy.signal import correlate2d

from transonic import boost

from .correl_pycuda import correl_pycuda
from .errors import PIVError
from .fft import CUFFT2DReal2Complex, FFTW2DReal2Complex, SKCUFFT2DReal2Complex
from .subpix import SubPix

try:
    import theano
except ImportError:
    pass


def compute_indices_from_displacement(dx, dy, indices_no_displ):
    return indices_no_displ[1] - dx, indices_no_displ[0] - dy


def parse_displacement_max(displ_max, im0_shape):
    if isinstance(displ_max, str) and displ_max.endswith("%"):
        return float(displ_max[:-1]) / 100 * max(im0_shape)

    else:
        return displ_max


def _compute_indices_max(correl, norm):

    iy, ix = np.unravel_index(np.nanargmax(correl), correl.shape)

    if norm == 0:
        # I hope it is ok (Pierre)
        correl_max = 0.0
    else:
        correl_max = correl[iy, ix] / norm

    if (
        iy == 0
        or iy == correl.shape[0] - 1
        or ix == 0
        or ix == correl.shape[1] - 1
    ):
        error = PIVError(explanation="Correlation peak touching boundary.")
        error.results = (ix, iy, correl_max)
        raise error

    if np.isnan(np.sum(correl[iy - 1 : iy + 2 : 2, ix - 1 : ix + 2 : 2])):
        error = PIVError(explanation="Correlation peak touching nan.")
        error.results = (ix, iy, correl_max)
        raise error

    return ix, iy, correl_max


class CorrelBase:
    """This class is meant to be subclassed, not instantiated directly."""

    _tag = "base"

    def __init__(
        self,
        im0_shape,
        im1_shape,
        method_subpix="centroid",
        nsubpix=1,
        displacement_max=None,
        particle_radius=3,
        nb_peaks_to_search=1,
        mode=None,
    ):

        self.mode = mode

        self.subpix = SubPix(method=method_subpix, nsubpix=nsubpix)

        self.im0_shape = im0_shape
        self.im1_shape = im1_shape
        self.iy0, self.ix0 = (i // 2 - 1 for i in im0_shape)

        self.displacement_max = parse_displacement_max(
            displacement_max, im0_shape
        )

        self.particle_radius = particle_radius
        self.nb_peaks_to_search = nb_peaks_to_search

        self._init2()

    def _init2(self):
        pass

    def compute_displacement_from_indices(self, ix, iy):
        """Compute the displacement from a couple of indices."""
        return self.ix0 - ix, self.iy0 - iy

    def compute_indices_from_displacement(self, dx, dy):
        return self.ix0 - dx, self.iy0 - dy

    def get_indices_no_displacement(self):
        return self.iy0, self.ix0

    def compute_displacements_from_correl(self, correl, norm=1.0):
        """Compute the displacement from a correlation."""

        try:
            ix, iy, correl_max = _compute_indices_max(correl, norm)
        except PIVError as e:
            ix, iy, correl_max = e.results
            # second chance to find a better peak...
            correl[
                iy - self.particle_radius : iy + self.particle_radius + 1,
                ix - self.particle_radius : ix + self.particle_radius + 1,
            ] = np.nan
            try:
                ix2, iy2, correl_max2 = _compute_indices_max(correl, norm)
            except PIVError as e2:
                dx, dy = self.compute_displacement_from_indices(ix, iy)
                e.results = (dx, dy, correl_max)
                raise e

            else:
                ix, iy, correl_max = ix2, iy2, correl_max2

        dx, dy = self.compute_displacement_from_indices(ix, iy)

        if self.nb_peaks_to_search == 1:
            other_peaks = None
        elif self.nb_peaks_to_search >= 1:
            other_peaks = []
            for ip in range(0, self.nb_peaks_to_search - 1):
                correl[
                    iy - self.particle_radius : iy + self.particle_radius + 1,
                    ix - self.particle_radius : ix + self.particle_radius + 1,
                ] = np.nan
                try:
                    ix, iy, correl_max_other = _compute_indices_max(correl, norm)
                except PIVError:
                    break

                dx_other, dy_other = self.compute_displacement_from_indices(
                    ix, iy
                )
                other_peaks.append((dx_other, dy_other, correl_max_other))
        # print('found {} peaks'.format(len(other_peaks) + 1))
        else:
            raise ValueError

        return dx, dy, correl_max, other_peaks

    def apply_subpix(self, dx, dy, correl):
        """Compute the displacement with the subpix method."""
        ix, iy = self.compute_indices_from_displacement(dx, dy)
        ix, iy = self.subpix.compute_subpix(correl, ix, iy)
        return self.compute_displacement_from_indices(ix, iy)


A = "float32[][]"


@boost
def correl_numpy(im0: A, im1: A, disp_max: int):
    """Correlations by hand using only numpy.

    Parameters
    ----------

    im0, im1 : images
      input images : 2D matrix

    disp_max : int
      displacement max.

    Notes
    -------

    im1_shape inf to im0_shape

    Returns
    -------

    the computing correlation (size of computed correlation = disp_max*2 + 1)

    """
    norm = np.sqrt(np.sum(im1**2) * np.sum(im0**2))

    ny = nx = int(disp_max) * 2 + 1
    ny0, nx0 = im0.shape
    ny1, nx1 = im1.shape

    zero = np.float32(0.0)
    correl = np.empty((ny, nx), dtype=np.float32)

    for xiy in range(disp_max + 1):
        dispy = -disp_max + xiy
        nymax = ny1 + min(ny0 // 2 - ny1 // 2 + dispy, 0)
        ny1dep = -min(ny0 // 2 - ny1 // 2 + dispy, 0)
        ny0dep = max(0, ny0 // 2 - ny1 // 2 + dispy)
        for xix in range(disp_max + 1):
            dispx = -disp_max + xix
            nxmax = nx1 + min(nx0 // 2 - nx1 // 2 + dispx, 0)
            nx1dep = -min(nx0 // 2 - nx1 // 2 + dispx, 0)
            nx0dep = max(0, nx0 // 2 - nx1 // 2 + dispx)
            tmp = zero
            for iy in range(nymax):
                for ix in range(nxmax):
                    tmp += (
                        im1[iy + ny1dep, ix + nx1dep]
                        * im0[ny0dep + iy, nx0dep + ix]
                    )
            correl[xiy, xix] = tmp / (nxmax * nymax)
        for xix in range(disp_max):
            dispx = xix + 1
            nxmax = nx1 - max(nx0 // 2 + nx1 // 2 + dispx - nx0, 0)
            nx1dep = 0
            nx0dep = nx0 // 2 - nx1 // 2 + dispx
            tmp = zero
            for iy in range(nymax):
                for ix in range(nxmax):
                    tmp += (
                        im1[iy + ny1dep, ix + nx1dep]
                        * im0[ny0dep + iy, nx0dep + ix]
                    )
            correl[xiy, xix + disp_max + 1] = tmp / (nxmax * nymax)
    for xiy in range(disp_max):
        dispy = xiy + 1
        nymax = ny1 - max(ny0 // 2 + ny1 // 2 + dispy - ny0, 0)
        ny1dep = 0
        ny0dep = ny0 // 2 - ny1 // 2 + dispy
        for xix in range(disp_max + 1):
            dispx = -disp_max + xix
            nxmax = nx1 + min(nx0 // 2 - nx1 // 2 + dispx, 0)
            nx1dep = -min(nx0 // 2 - nx1 // 2 + dispx, 0)
            nx0dep = max(0, nx0 // 2 - nx1 // 2 + dispx)
            tmp = zero
            for iy in range(nymax):
                for ix in range(nxmax):
                    tmp += (
                        im1[iy + ny1dep, ix + nx1dep]
                        * im0[ny0dep + iy, nx0dep + ix]
                    )
            correl[xiy + disp_max + 1, xix] = tmp / (nxmax * nymax)
        for xix in range(disp_max):
            dispx = xix + 1
            nxmax = nx1 - max(nx0 // 2 + nx1 // 2 + dispx - nx0, 0)
            nx1dep = 0
            nx0dep = nx0 // 2 - nx1 // 2 + dispx
            tmp = zero
            for iy in range(nymax):
                for ix in range(nxmax):
                    tmp += (
                        im1[iy + ny1dep, ix + nx1dep]
                        * im0[ny0dep + iy, nx0dep + ix]
                    )
            correl[xiy + disp_max + 1, xix + disp_max + 1] = tmp / (nxmax * nymax)
    correl = correl * im1.size
    return correl, norm


class CorrelPythran(CorrelBase):
    """Correlation using pythran.
    Correlation class by hands with with numpy.
    """

    _tag = "pythran"

    def _init2(self):

        if self.displacement_max is None:
            if self.im0_shape == self.im1_shape:
                displacement_max = min(self.im0_shape) // 2 - 1
            else:
                displacement_max = (
                    min(
                        self.im0_shape[0] - self.im1_shape[0],
                        self.im0_shape[1] - self.im1_shape[1],
                    )
                    // 2
                    - 1
                )
        if displacement_max <= 0:
            raise ValueError(
                "displacement_max <= 0 : problem with images shapes?"
            )

        self.displacement_max = displacement_max

        self.ix0 = displacement_max
        self.iy0 = displacement_max

    def __call__(self, im0, im1):
        """Compute the correlation from images."""
        correl, norm = correl_numpy(im0, im1, self.displacement_max)
        return correl, norm


class CorrelPyCuda(CorrelBase):
    """Correlation using pycuda.
    Correlation class by hands with with cuda.
    """

    _tag = "pycuda"

    def _init2(self):

        if self.displacement_max is None:
            if self.im0_shape == self.im1_shape:
                displacement_max = min(self.im0_shape) // 2
            else:
                displacement_max = (
                    min(
                        self.im0_shape[0] - self.im1_shape[0],
                        self.im0_shape[1] - self.im1_shape[1],
                    )
                    // 2
                    - 1
                )
        if displacement_max <= 0:
            raise ValueError(
                "displacement_max <= 0 : problem with images shapes?"
            )

        self.displacement_max = displacement_max

        self.ix0 = displacement_max
        self.iy0 = displacement_max

    def __call__(self, im0, im1):
        """Compute the correlation from images."""
        correl, norm = correl_pycuda(im0, im1, self.displacement_max)
        self._add_info_to_correl(correl)
        return correl, norm


class CorrelScipySignal(CorrelBase):
    """Correlations using scipy.signal.correlate2d"""

    _tag = "scipy.signal"

    def _init2(self):
        if self.mode is None:
            self.mode = "same"

        modes = ["valid", "same"]
        if self.mode not in modes:
            raise ValueError("mode should be in " + modes)

        if self.mode == "same":
            ny, nx = self.im0_shape
            if nx % 2 == 0:
                ind0x = nx // 2 - 1
            else:
                ind0x = nx // 2
            if ny % 2 == 0:
                ind0y = ny // 2 - 1
            else:
                ind0y = ny // 2

        else:
            ny, nx = np.array(self.im0_shape) - np.array(self.im1_shape)
            ind0x = nx // 2
            ind0y = ny // 2

        self.iy0, self.ix0 = (ind0y, ind0x)

    def __call__(self, im0, im1):
        """Compute the correlation from images."""
        norm = np.sqrt(np.sum(im1**2) * np.sum(im0**2))
        if self.mode == "valid":
            correl = correlate2d(im0, im1, mode="valid")
        elif self.mode == "same":
            correl = correlate2d(im0, im1, mode="same", fillvalue=im1.min())
        else:
            assert False, "Bad value for self.mode"

        return correl, norm


class CorrelScipyNdimage(CorrelBase):
    """Correlations using scipy.ndimage.correlate."""

    _tag = "scipy.ndimage"

    def _init2(self):
        self.iy0, self.ix0 = (i // 2 for i in self.im0_shape)

    def __call__(self, im0, im1):
        """Compute the correlation from images."""
        norm = np.sum(im1**2)
        correl = correlate(im0, im1, mode="constant", cval=im1.min())

        return correl, norm


class CorrelTheano(CorrelBase):
    """Correlations using theano.tensor.nnet.conv2d"""

    _tag = "theano"

    def _init2(self):
        if self.mode is None:
            mode = self.mode = "disp"

        im0_shape = self.im0_shape
        im1_shape = self.im1_shape

        if im1_shape is None:
            im1_shape = im0_shape

        if self.displacement_max is None:
            if im0_shape == im1_shape:
                displacement_max = max(im0_shape) // 2
            else:
                displacement_max = (
                    max(im0_shape[0] - im1_shape[0], im0_shape[1] - im1_shape[1])
                    // 2
                    - 1
                )
        if displacement_max <= 0:
            raise ValueError("displacement_max <= 0 : problem with images shapes")

        modes = ["valid", "same", "disp"]
        if mode not in modes:
            raise ValueError("mode should be in " + modes)

        self.mode = mode
        self.ny0, self.nx0 = im0_shape
        self.ny1, self.nx1 = im1_shape
        self.displacement_max = displacement_max
        if mode == "same":
            self.ny, self.nx = im0_shape
            if self.nx % 2 == 0:
                ind0x = self.nx // 2 - 1
            else:
                ind0x = self.nx // 2
            if self.ny % 2 == 0:
                ind0y = self.ny // 2 - 1
            else:
                ind0y = self.ny // 2

        elif mode == "valid":
            self.ny, self.nx = np.array(im0_shape) - np.array(im1_shape) + 1
            ind0x = self.nx // 2
            ind0y = self.ny // 2
        else:
            self.ny = displacement_max * 2 + 1
            self.nx = self.ny
            ind0x = displacement_max
            ind0y = displacement_max

        im00 = theano.tensor.tensor4("im00", dtype="float32")
        im11 = theano.tensor.tensor4("im11", dtype="float32")
        modec = theano.compile.get_default_mode()
        # modec = modec.including('conv_meta')
        if mode == "same":
            correl_theano = theano.tensor.nnet.conv2d(
                im00,
                im11,
                image_shape=(1, 1, 2 * self.ny0 - 1, 2 * self.nx0 - 1),
                filter_shape=(1, 1) + im1_shape,
                border_mode="valid",
            )
        elif mode == "valid":
            correl_theano = theano.tensor.nnet.conv2d(
                im00,
                im11,
                input_shape=(1, 1) + im0_shape,
                filter_shape=(1, 1) + im1_shape,
                border_mode="valid",
            )
        else:
            if (self.ny0 <= 2 * self.displacement_max + self.ny1) & (
                self.nx0 <= 2 * self.displacement_max + self.nx1
            ):
                correl_theano = theano.tensor.nnet.conv2d(
                    im00,
                    im11,
                    input_shape=(
                        1,
                        1,
                        2 * displacement_max + self.ny1,
                        2 * displacement_max + self.nx1,
                    ),
                    filter_shape=(1, 1) + im1_shape,
                    border_mode="valid",
                )
            elif (self.ny0 > 2 * self.displacement_max + self.ny1) & (
                self.nx0 > 2 * self.displacement_max + self.nx1
            ):
                correl_theano = theano.tensor.nnet.conv2d(
                    im00,
                    im11,
                    image_shape=(1, 1) + im0_shape,
                    filter_shape=(1, 1) + im1_shape,
                    border_mode="valid",
                )
            else:
                assert False, "Bad value for self.mode"

        self.correlf = theano.function(
            inputs=[im00, im11], outputs=[correl_theano], mode=modec
        )

        self.iy0, self.ix0 = (ind0y, ind0x)

    def __call__(self, im0, im1):
        """Compute the correlation from images."""
        norm = np.sqrt(np.sum(im1**2) * np.sum(im0**2))
        im1 = np.rot90(im1, 2)
        im1 = im1.reshape(1, 1, self.ny1, self.nx1)
        if self.mode == "valid":
            im0 = im0.reshape(1, 1, self.ny0, self.nx0)
        elif self.mode == "same":
            im0b = im1.min() * np.ones(
                (2 * self.ny - 1, 2 * self.nx - 1), dtype=np.float32
            )
            im0b[
                self.ny // 2 - 1 : self.ny + self.ny // 2 - 1,
                self.nx // 2 - 1 : self.nx + self.nx // 2 - 1,
            ] = im0
            # Correlation with periodic condition (==FFT version) :
            # im0 = np.tile(im0, (3, 3))
            # im0 = im0[self.nx//2+1:2*self.nx+self.nx//2,
            #           self.ny//2+1:2*self.ny+self.ny//2]
            im0 = im0b.reshape(1, 1, 2 * self.ny - 1, 2 * self.nx - 1)
        elif self.mode == "disp":
            if (self.ny0 < 2 * self.displacement_max + self.ny1) & (
                self.nx0 < 2 * self.displacement_max + self.nx1
            ):

                im0b = np.zeros(
                    (
                        2 * self.displacement_max + self.ny1,
                        2 * self.displacement_max + self.nx1,
                    ),
                    dtype=np.float32,
                )
                i00 = (2 * self.displacement_max + self.nx1 - self.nx0) // 2
                j00 = (2 * self.displacement_max + self.ny1 - self.ny0) // 2
                im0b[j00 : self.ny0 + j00, i00 : self.nx0 + i00] = im0
                im0 = im0b.reshape(
                    1,
                    1,
                    2 * self.displacement_max + self.ny1,
                    2 * self.displacement_max + self.nx1,
                )
            elif (self.ny0 > 2 * self.displacement_max + self.ny1) & (
                self.nx0 > 2 * self.displacement_max + self.nx1
            ):
                im0 = im0.reshape(1, 1, self.ny0, self.nx0)
        else:
            assert False, "Bad value for self.mode"

        correl = self.correlf(im0, im1)
        correl = np.asarray(correl)
        if (
            (self.ny0 > 2 * self.displacement_max + self.ny1)
            & (self.nx0 > 2 * self.displacement_max + self.nx1)
            & (self.mode == "disp")
        ):
            i00 = (self.nx0 - self.nx1 + 1) // 2 - self.displacement_max
            j00 = (self.ny0 - self.ny1 + 1) // 2 - self.displacement_max
            correl = correl[
                0,
                0,
                0,
                j00 : j00 + 2 * self.displacement_max + 1,
                i00 : i00 + 2 * self.displacement_max + 1,
            ]
        else:
            correl = correl.reshape(self.ny, self.nx)

        return correl, norm


class CorrelFFTBase(CorrelBase):
    """Correlations using fft."""

    _tag = "fft.base"

    def _init2(self):

        if self.displacement_max is not None:
            where_large_displacement = np.zeros(self.im0_shape, dtype=bool)

            for indices, v in np.ndenumerate(where_large_displacement):
                dx, dy = self.compute_displacement_from_indices(*indices[::-1])
                displacement = np.sqrt(dx**2 + dy**2)
                if displacement > self.displacement_max:
                    where_large_displacement[indices] = True

            self.where_large_displacement = where_large_displacement

        self._check_im_shape(self.im0_shape, self.im1_shape)

    def _check_im_shape(self, im0_shape, im1_shape):
        if im1_shape is None:
            im1_shape = im0_shape

        if im0_shape != im1_shape:
            raise ValueError(
                "with this correlation method the input images "
                "have to have the same shape."
            )

    def compute_displacements_from_correl(self, correl, norm=1.0):

        """Compute the displacement (with subpix) from a correlation."""

        if self.displacement_max is not None:
            correl = correl.copy()
            correl[self.where_large_displacement] = np.nan

        return super().compute_displacements_from_correl(correl, norm=norm)


class CorrelFFTNumpy(CorrelFFTBase):
    """Correlations using numpy.fft."""

    _tag = "np.fft"

    def __call__(self, im0, im1):
        """Compute the correlation from images."""
        norm = np.sqrt(np.sum(im1**2) * np.sum(im0**2))
        corr = ifft2(fft2(im0).conj() * fft2(im1)).real
        correl = np.fft.fftshift(corr[::-1, ::-1])
        return correl, norm


class CorrelFFTW(CorrelFFTBase):
    """Correlations using fluidimage.fft.FFTW2DReal2Complex"""

    FFTClass = FFTW2DReal2Complex
    _tag = "fftw"

    def _init2(self):
        CorrelFFTBase._init2(self)
        n0, n1 = self.im0_shape
        self.op = self.FFTClass(n1, n0)

    def __call__(self, im0, im1):
        """Compute the correlation from images."""
        norm = np.sqrt(np.sum(im1**2) * np.sum(im0**2)) * im0.size
        op = self.op
        corr = op.ifft(op.fft(im0).conj() * op.fft(im1))
        correl = np.fft.fftshift(corr[::-1, ::-1])
        return correl, norm


class CorrelCuFFT(CorrelFFTBase):
    _tag = "cufft"
    """Correlations using fluidimage.fft.CUFFT2DReal2Complex"""
    FFTClass = CUFFT2DReal2Complex

    def _init2(self):
        CorrelFFTBase._init2(self)
        n0, n1 = self.im0_shape
        self.op = self.FFTClass(n1, n0)

    def __call__(self, im0, im1):
        """Compute the correlation from images."""
        norm = np.sqrt(np.sum(im1**2) * np.sum(im0**2)) * im0.size
        op = self.op
        corr = op.ifft(op.fft(im0).conj() * op.fft(im1)).real * im0.size**2
        correl = np.fft.fftshift(corr[::-1, ::-1])
        return correl, norm


class CorrelSKCuFFT(CorrelFFTBase):
    """Correlations using fluidimage.fft.FFTW2DReal2Complex"""

    FFTClass = SKCUFFT2DReal2Complex
    _tag = "skcufft"

    def _init2(self):
        CorrelFFTBase._init2(self)
        n0, n1 = self.im0_shape
        self.op = self.FFTClass(n1, n0)

    def __call__(self, im0, im1):
        """Compute the correlation from images."""
        norm = np.sqrt(np.sum(im1**2) * np.sum(im0**2)) * im0.size
        op = self.op
        corr = op.ifft(op.fft(im0).conj() * op.fft(im1))
        correl = np.fft.fftshift(corr[::-1, ::-1])
        return correl, norm


correlation_classes = {
    v._tag: v
    for k, v in locals().items()
    if k.startswith("Correl") and not k.endswith("Base")
}



import sys
from pathlib import Path

import numpy as np

from fluidimage.util import imread, logger

try:
    import pyqtgraph as pg
    from pyqtgraph import QtCore, QtGui, dockarea
except ImportError:
    from warnings import warn

    warn(
        "PyQtgraph must be installed to use this module! "
        + "Also one of these must be installed: PyQt>=4.7 / PySide / PyQt5"
    )


class PGWrapper:
    def __init__(self, win_type=None, title="FluidImage"):
        self._create_app()
        self._create_win(win_type, title)
        self._type = win_type
        self.area = None

    def _create_app(self):
        self.app = QtGui.QApplication([])

    def _create_win(self, win_type, title):
        if win_type is None:
            self.win = QtGui.QMainWindow()
            self.win.resize(800, 800)
        elif win_type == "glt_widget":
            self.win = pg.GraphicsLayoutWidget()
        elif win_type == "gfx_window":
            self.win = pg.GraphicsWindow()

        self.win.setWindowTitle(title)

    def _add_dock(self, title="Dock", size=(500, 200), position="bottom"):
        if self.area is None:
            self.area = dockarea.DockArea()
            self.win.setCentralWidget(self.area)

        d = dockarea.Dock(title, size=size)
        self.area.addDock(d, position)
        logger.debug("Function _add_dock: " + repr(d))

    def _add_gfx_item(self, obj, item):
        # GWidget = pg.graphicsItems.GraphicsWidget.GraphicsWidget
        GPlotItem = pg.graphicsItems.GraphicsLayout.PlotItem
        GItem = pg.graphicsItems.GraphicsItem.GraphicsItem

        attr = " "
        if hasattr(obj, "addPlot") and isinstance(item, GPlotItem):
            obj.addPlot(item)
            attr = "addPlot"
        elif hasattr(obj, "addItem") and isinstance(item, GItem):
            obj.addItem(item)
            attr = "addItem"
        elif hasattr(obj, "addWidget"):
            obj.addWidget(item)
            attr = "addWidget"
        elif hasattr(obj, "setCentralWidget"):
            obj.setCentralWidget(item)
            attr = "setCentralWidget"

        logger.debug("Function _add_gfx_item: " + attr + repr(item))

    def _win(self, key=None, attr="__init__"):
        if self.area is None:
            win = self.win
        elif key is None:
            win = self.area.docks.items()[0][1]
        else:
            win = self.area.docks[key]

        logger.debug("Function _win: " + repr(win))
        return win

    def view(self, path, title=None, hide_crosshair=True):
        """
        ImageView, a high-level widget for displaying and analyzing 2D and 3D
        data. ImageView provides:

          1. A zoomable region (ViewBox) for displaying the image
          2. A combination histogram and gradient editor (HistogramLUTItem)
             for controlling the visual appearance of the image
          3. A timeline for selecting the currently displayed frame (for 3D
             data only).
          4. Tools for very basic analysis of image data (see ROI and Norm
             buttons)

        """
        imv = pg.ImageView()
        win = self._win(title)
        self._add_gfx_item(win, imv)

        if not isinstance(path, str):
            data = []
            for p in path:
                logger.info(f"Viewing {p}")
                data.append(imread(p).transpose())

            data = np.array(data)
            imv.setImage(data, xvals=np.linspace(0, len(path), data.shape[0]))
        elif Path(path).is_dir():
            raise ValueError("Expected files not directory.")
        else:
            logger.info(f"Viewing {path}")
            try:
                data = imread(path).transpose()
            except AttributeError:
                raise ValueError(f"Is {path} an image?")
            imv.setImage(data)

        vb = imv.imageItem.getViewBox()
        self._add_crosshair(win, imv, vb, hide_lines=hide_crosshair)

    def _add_crosshair(self, win, p1, vb, hide_lines=True):
        if not hide_lines:
            vLine = pg.InfiniteLine(angle=90, movable=False)
            hLine = pg.InfiniteLine(angle=0, movable=False)
            p1.addItem(vLine, ignoreBounds=True)
            p1.addItem(hLine, ignoreBounds=True)

        label = pg.TextItem("(x,y)=intensity", anchor=(0.0, 0.0))
        p1.addItem(label)
        label.hide()

        def mouseMoved(evt):
            data1 = (
                p1.getImageItem().image
            )  # should be updated using sigTimeChanged
            pos = evt[
                0
            ]  # using signal proxy turns original arguments into a tuple
            mousePoint = vb.mapSceneToView(pos)
            x = mousePoint.x()
            y = mousePoint.y()
            index = np.array([x, y], dtype=int)
            index_max = data1.shape
            if np.all(np.greater_equal(index, [0, 0])) and np.all(
                np.less_equal(index, index_max)
            ):
                text = (
                    "<span style='font-size: 10pt; color: cyan'>(%0.1f, %0.1f)=</span>"
                    "<span style='color: red'>%0.1f</span>"
                ) % (mousePoint.x(), mousePoint.y(), data1[index[0], index[1]])
                label.setHtml(text)
                label.setPos(x, y)
                label.show()
                if not hide_lines:
                    vLine.setPos(x)
                    hLine.setPos(y)
            else:
                label.hide()

        try:
            signal = p1.scene.sigMouseMoved
        except AttributeError:
            signal = p1.scene().sigMouseMoved

        self.proxy = pg.SignalProxy(signal, rateLimit=48, slot=mouseMoved)

    def show(self):
        self.win.show()
        if (sys.flags.interactive != 1) or not hasattr(QtCore, "PYQT_VERSION"):
            QtGui.QApplication.instance().exec_()
        else:
            raise ValueError("Cannot start Qt event while in interactive mode")



"""Tomographic Reconstruction algorithms
========================================

Provides:

.. autosummary::
   :toctree:

   mlos

.. autoclass:: TomoMLOSCV
   :members:
   :private-members:


"""
from .mlos import TomoMLOSCV



"""Executor async/await (:mod:`fluidimage.executors.exec_async`)
================================================================

This executer uses await/async with trio library to put topology tasks in
concurrency.

A single executor (in one process) is used. If CPU bounded tasks are limited by
the Python GIL, the threads won't use at the same time the CPU.

This means that the work will run on one CPU at a time, except if the topology
uses compiled code releasing the GIL. In this case, the GIL can be bypassed and
computation can use many CPU at a time.

.. autoclass:: ExecutorAsync
   :members:
   :private-members:

"""

import signal
import time
from collections import OrderedDict

import trio

from fluidimage.util import cstring, log_debug, log_memory_usage, logger

from .base import ExecutorBase


class ExecutorAsync(ExecutorBase):
    """Executor async/await.

    The work in performed in a single process.

    """

    def __init__(
        self,
        topology,
        path_dir_result,
        nb_max_workers=None,
        nb_items_queue_max=None,
        sleep_time=0.01,
        logging_level="info",
        stop_if_error=False,
    ):
        super().__init__(
            topology,
            path_dir_result,
            nb_max_workers,
            nb_items_queue_max,
            logging_level=logging_level,
            stop_if_error=stop_if_error,
        )

        self.nb_working_workers_cpu = 0
        self.nb_working_workers_io = 0

        # Executor parameters
        self.sleep_time = sleep_time

        # function containers
        self.async_funcs = OrderedDict()
        self.funcs = OrderedDict()

        # Functions definition
        self.define_functions()

        def signal_handler(sig, frame):
            del sig, frame  # unused
            logger.info("Ctrl+C signal received...")
            self._has_to_stop = True
            self.nursery.cancel_scope.cancel()
            # it seems that we don't need to raise the exception
            raise KeyboardInterrupt

        signal.signal(signal.SIGINT, signal_handler)

        # to avoid a pylint warning
        self.nursery = None

    def compute(self):
        """Compute the whole topology.

        Begin by executing one shot jobs, then execute multiple shots jobs
        implemented as async functions.  Warning, one shot jobs must be ancestors
        of multiple shots jobs in the topology.

        """

        self._init_compute()
        self.exec_one_shot_works()
        trio.run(self.start_async_works)
        self._finalize_compute()

    async def start_async_works(self):
        """Create a trio nursery and start all async functions."""
        async with trio.open_nursery() as self.nursery:
            for af in self.async_funcs.values():
                self.nursery.start_soon(af)

            self.nursery.start_soon(self.update_has_to_stop)

    def define_functions(self):
        """Define sync and async functions.

        Define sync ("one shot" functions) and async functions (multiple shot
        functions), and store them in `self.async_funcs`.

        The behavior of the executor is mostly defined here.  To sum up : Each
        "multiple shot" waits for an items to be available in there input_queue
        and process the items as soon as they are available.

        """
        for work in self.works:

            # global functions
            if work.kind is not None and "global" in work.kind:

                async def func(work=work):
                    while True:
                        while (
                            isinstance(work.input_queue, tuple)
                            and all(not q for q in work.input_queue)
                        ) or not work.input_queue:
                            await trio.sleep(self.sleep_time)
                            if self._has_to_stop:
                                return
                        t_start = time.time()
                        log_memory_usage(
                            f"{time.time() - self.t_start:.2f} s. Launch work "
                            + work.name_no_space
                            + f" (?). mem usage"
                        )
                        work.func_or_cls(work.input_queue, work.output_queue)
                        if self._has_to_stop:
                            return
                        await trio.sleep(self.sleep_time)

                        logger.info(
                            f"work {work.name_no_space} "
                            f"done in {time.time() - t_start:.3f} s"
                        )

                        await trio.sleep(self.sleep_time)

            # I/O
            elif work.kind is not None and (
                "io" in work.kind or work.kind == "io"
            ):
                func = self.def_async_func_work_io(work)

            # CPU-bounded work
            else:
                func = self.def_async_func_work_cpu(work)

            self.async_funcs[work.name] = func

    def def_async_func_work_io(self, work):
        """Define an asynchronous function launching a io work."""

        async def func(work=work):
            while True:
                while (
                    not work.input_queue
                    or self.nb_working_workers_io >= self.nb_max_workers
                    or (
                        work.output_queue is not None
                        and len(work.output_queue) >= self.nb_items_queue_max
                    )
                ):
                    if self._has_to_stop:
                        return
                    await trio.sleep(self.sleep_time)
                self.nursery.start_soon(self.async_run_work_io, work)
                await trio.sleep(self.sleep_time)

        return func

    def def_async_func_work_cpu(self, work):
        """Define an asynchronous function launching a cpu work."""

        async def func(work=work):
            while True:
                while (
                    not work.input_queue
                    or self.nb_working_workers_cpu >= self.nb_max_workers
                    or (
                        work.output_queue is not None
                        and len(work.output_queue) >= self.nb_items_queue_max
                    )
                ):
                    if self._has_to_stop:
                        return
                    await trio.sleep(self.sleep_time)
                self.nursery.start_soon(self.async_run_work_cpu, work)
                await trio.sleep(self.sleep_time)

        return func

    async def async_run_work_io(self, work):
        """Is destined to be started with a "trio.start_soon".

        Executes the work on an item (key, obj), and add the result on
        work.output_queue.

        Parameters
        ----------

        work :

          A work from the topology

        """
        self.nb_working_workers_io += 1

        try:
            key, obj = work.input_queue.pop_first_item()
        except KeyError:
            self.nb_working_workers_io -= 1
            return

        if work.check_exception(key, obj):
            self.nb_working_workers_io -= 1
            return

        t_start = time.time()
        log_memory_usage(
            f"{time.time() - self.t_start:.2f} s. Launch work "
            + work.name_no_space
            + f" ({key}). mem usage"
        )
        # pylint: disable=W0703
        try:
            ret = await trio.to_thread.run_sync(work.func_or_cls, obj)
        except Exception as error:
            self.log_exception(error, work.name_no_space, key)
            if self.stop_if_error:
                raise
            ret = error
        else:
            logger.info(
                f"work {work.name_no_space} ({key}) "
                f"done in {time.time() - t_start:.3f} s"
            )

        if work.output_queue is not None:
            work.output_queue[key] = ret
        self.nb_working_workers_io -= 1

    async def async_run_work_cpu(self, work):
        """Is destined to be started with a "trio.start_soon".

        Executes the work on an item (key, obj), and add the result on
        work.output_queue.

        Parameters
        ----------

        work :

          A work from the topology

        """
        self.nb_working_workers_cpu += 1

        try:
            key, obj = work.input_queue.pop_first_item()
        except KeyError:
            self.nb_working_workers_cpu -= 1
            return

        if work.check_exception(key, obj):
            self.nb_working_workers_cpu -= 1
            return

        t_start = time.time()
        log_memory_usage(
            f"{time.time() - self.t_start:.2f} s. Launch work "
            + work.name_no_space
            + f" ({key}). mem usage"
        )
        # pylint: disable=W0703
        try:
            ret = await trio.to_thread.run_sync(work.func_or_cls, obj)
        except Exception as error:
            self.log_exception(error, work.name_no_space, key)
            if self.stop_if_error:
                raise
            ret = error
        else:
            logger.info(
                f"work {work.name_no_space} ({key}) "
                f"done in {time.time() - t_start:.3f} s"
            )

        if work.output_queue is not None:
            work.output_queue[key] = ret
        self.nb_working_workers_cpu -= 1

    async def update_has_to_stop(self):
        """Work has to stop flag. Check if all works has been done.

        Return True if there are no workers in working and if there is no items in
        all queues.

        """

        while not self._has_to_stop:

            result = (
                (not any([bool(queue) for queue in self.topology.queues]))
                and self.nb_working_workers_cpu == 0
                and self.nb_working_workers_io == 0
            )

            if result:
                self._has_to_stop = True
                log_debug(f"has_to_stop!")

            if self.logging_level == "debug":
                log_debug(f"self.topology.queues: {self.topology.queues}")
                log_debug(
                    f"self.nb_working_workers_cpu: {self.nb_working_workers_cpu}"
                )
                log_debug(
                    f"self.nb_working_workers_io: {self.nb_working_workers_io}"
                )

            await trio.sleep(self.sleep_time)



import unittest
from pathlib import Path
from shutil import rmtree

from fluidimage import get_path_image_samples
from fluidimage.topologies.optical_flow import TopologyOpticalFlow

path_image_samples = get_path_image_samples()


class TestPivNew(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.path_Oseen = path_image_samples / "Oseen/Images/Oseen*"
        cls.path_Jet = path_image_samples / "Jet/Images/c06*"
        cls.postfix = "test_optical_flow"

    @classmethod
    def tearDownClass(cls):
        paths = (cls.path_Oseen, cls.path_Jet)
        for path in paths:
            path_out = Path(str(path.parent) + "." + cls.postfix)
            if path_out.exists():
                rmtree(path_out)

    def test_optical_flow(self):
        params = TopologyOpticalFlow.create_default_params()

        params.series.path = str(self.path_Oseen)
        params.series.ind_start = 1
        params.series.ind_step = 1

        # params.saving.how has to be equal to 'complete' for idempotent jobs
        # (on clusters)
        params.saving.how = "recompute"
        params.saving.postfix = self.postfix

        params.filters.displacement_max = 10.0

        topology = TopologyOpticalFlow(params, logging_level="info")

        topology.make_code_graphviz(topology.path_dir_result / "topo.dot")
        topology.compute("exec_sequential")


if __name__ == "__main__":
    unittest.main()



import matplotlib.pyplot as plt
import numpy as np

from fluidimage.calcul.correl import CorrelScipySignal
from fluidimage.synthetic import make_synthetic_images

plt.ion()


nx = 64
ny = 64

nx1 = 32
ny1 = 32

displacement_x = 2.0
displacement_y = 2.0

displacements = np.array([displacement_y, displacement_x])

nb_particles = (nx // 3) ** 2


print(f"nx: {nx} ; ny: {ny} ; nx1: {nx1} ; ny1: {ny1}")

im0, im1 = make_synthetic_images(
    displacements,
    nb_particles,
    shape_im0=(ny, nx),
    shape_im1=(ny1, nx1),
    epsilon=0.0,
)

# plt.figure()

# ax0 = plt.subplot(121)
# ax1 = plt.subplot(122)

# axi0 = ax0.imshow(im0, interpolation='nearest')
# axi1 = ax1.imshow(im1, interpolation='nearest')


classes = {"sig": CorrelScipySignal}


cs = {}
funcs = {}
for k, cls in classes.items():
    calcul_corr = cls(im0.shape, im1.shape, mode="valid")
    funcs[k] = calcul_corr
    cs[k], norm = calcul_corr(im0, im1)


for k, c in cs.items():
    func = funcs[k]
    inds_max = np.array(np.unravel_index(c.argmax(), c.shape))
    if not np.allclose(
        displacements.astype("int"),
        func.compute_displacement_from_indices(inds_max),
    ):
        print(
            "do not understand " + k,
            displacements.astype("int"),
            func.compute_displacement_from_indices(inds_max),
        )

c = cs["sig"]
# plt.figure()

# ax = plt.gca()
# ax.imshow(c, interpolation='nearest')

# plt.show()



import os

import numpy as np

try:
    import pycuda._driver
except ImportError:
    pass
else:
    try:
        import pycuda.autoinit
        import pycuda.compiler
        import pycuda.gpuarray
        import pycuda.driver
    except (ImportError, pycuda._driver.RuntimeError):
        pass


def nextpow2(i):
    n = 1
    while n < i:
        n *= 2
    return n


def correl_pycuda(im0, im1, disp_max):
    """Correlations by hand using only numpy.

    Parameters
    ----------

    im0, im1 : images
      input images : 2D matrix

    disp_max : int
      displacement max.

    Notes
    -------

    im1_shape inf to im0_shape

    Returns
    -------

    the computing correlation (size of computed correlation = disp_max*2 + 1)

    """
    norm = np.sqrt(np.sum(im1**2) * np.sum(im0**2)) * im0.size

    ny = nx = int(disp_max) * 2 + 1
    ny0, nx0 = im0.shape
    ny1, nx1 = im1.shape

    # zero = np.float32(0.)
    correl = np.empty((ny, nx), dtype=np.float32)

    # Load the kernel and compile it.
    kernel_file = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "correl.cu"
    )
    with open(kernel_file, "r") as file:
        kernel = pycuda.compiler.SourceModule(file.read())
    correlate_cuda = kernel.get_function("cucorrelate")
    # correlate_cuda.prepare([np.intp, np.intp, np.intp, np.int32,
    #                        np.int32,np.int32,np.int32,np.int32,np.int32,
    #                        np.int32,np.int32,np.int32,np.int32])

    #    correlate_cuda.prepare([numpy.intp,numpy.intp,numpy.intp,numpy.int32,numpy.int32,
    #                    numpy.int32,numpy.int32,numpy.int32,numpy.int32,
    #                    numpy.int32,numpy.int32,numpy.int32,numpy.int32],block=(nthreads_x,nthreads_x,nthreads_x))
    kernel_file = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "reduction_kernel.cu"
    )
    with open(kernel_file, "r") as file:
        kernel1 = pycuda.compiler.SourceModule(file.read())

    reduction_cuda = kernel1.get_function("reduce1")

    # CUDA parameters that seem to work well. The number of threads per tile
    # (the tile_size) should be a power of 2 for the parallel reduction to
    # work right!
    threads_per_block = 16
    blocks_per_tile = max(nx0, ny0) / threads_per_block

    maxthreads = 16

    g_im0data = pycuda.gpuarray.to_gpu(im0)
    g_im1data = pycuda.gpuarray.to_gpu(im1)

    g_odata = pycuda.gpuarray.empty(nx0 * ny0, np.float32)
    g_sumdata = pycuda.gpuarray.empty(1, np.float32)

    for xiy in range(disp_max + 1):
        dispy = -disp_max + xiy
        nymax = ny1 + min(ny0 // 2 - ny1 // 2 + dispy, 0)
        ny1dep = -min(ny0 // 2 - ny1 // 2 + dispy, 0)
        ny0dep = max(0, ny0 // 2 - ny1 // 2 + dispy)
        for xix in range(disp_max + 1):
            dispx = -disp_max + xix
            nxmax = nx1 + min(nx0 // 2 - nx1 // 2 + dispx, 0)
            nx1dep = -min(nx0 // 2 - nx1 // 2 + dispx, 0)
            nx0dep = max(0, nx0 // 2 - nx1 // 2 + dispx)

            threads_reduce = maxthreads
            blocks_reduce = nx0 * ny0
            if blocks_reduce < maxthreads:
                threads_reduce = nextpow2(blocks_reduce)
            blocks_reduce = (blocks_reduce + threads_reduce - 1) // threads_reduce
            correlate_cuda(
                g_im0data.gpudata,
                g_im1data.gpudata,
                g_odata.gpudata,
                np.int32(nx0),
                np.int32(nx1),
                np.int32(ny0),
                np.int32(dispx),
                np.int32(dispy),
                np.int32(nymax),
                np.int32(ny1dep),
                np.int32(ny0dep),
                np.int32(nx0dep),
                np.int32(nxmax),
                np.int32(nx1dep),
                block=(threads_per_block, threads_per_block, 1),
                grid=(blocks_per_tile, blocks_per_tile),
            )
            while blocks_reduce != 1:
                reduction_cuda(
                    g_odata.gpudata,
                    g_odata.gpudata,
                    block=(threads_reduce, 1, 1),
                    grid=(blocks_reduce, 1),
                    shared=4 * threads_reduce,
                )
                if blocks_reduce < maxthreads:
                    threads_reduce = nextpow2(blocks_reduce)
                blocks_reduce = (
                    blocks_reduce + threads_reduce - 1
                ) // threads_reduce
            reduction_cuda(
                g_odata.gpudata,
                g_sumdata.gpudata,
                block=(threads_reduce, 1, 1),
                grid=(blocks_reduce, 1),
                shared=4 * threads_reduce,
            )

            correl[xiy, xix] = g_sumdata.get()

        for xix in range(disp_max):
            dispx = xix + 1
            nxmax = nx1 - max(nx0 // 2 + nx1 // 2 + dispx - nx0, 0)
            nx1dep = 0
            nx0dep = nx0 // 2 - nx1 // 2 + dispx

            threads_reduce = maxthreads
            blocks_reduce = nx0 * ny0
            if blocks_reduce < maxthreads:
                threads_reduce = nextpow2(blocks_reduce)
            blocks_reduce = (blocks_reduce + threads_reduce - 1) // threads_reduce
            correlate_cuda(
                g_im0data.gpudata,
                g_im1data.gpudata,
                g_odata.gpudata,
                np.int32(nx0),
                np.int32(nx1),
                np.int32(ny0),
                np.int32(dispx),
                np.int32(dispy),
                np.int32(nymax),
                np.int32(ny1dep),
                np.int32(ny0dep),
                np.int32(nx0dep),
                np.int32(nxmax),
                np.int32(nx1dep),
                block=(threads_per_block, threads_per_block, 1),
                grid=(blocks_per_tile, blocks_per_tile),
            )
            while blocks_reduce != 1:
                reduction_cuda(
                    g_odata.gpudata,
                    g_odata.gpudata,
                    block=(threads_reduce, 1, 1),
                    grid=(blocks_reduce, 1),
                    shared=4 * threads_reduce,
                )
                if blocks_reduce < maxthreads:
                    threads_reduce = nextpow2(blocks_reduce)
                blocks_reduce = (
                    blocks_reduce + threads_reduce - 1
                ) // threads_reduce
            reduction_cuda(
                g_odata.gpudata,
                g_sumdata.gpudata,
                block=(threads_reduce, 1, 1),
                grid=(blocks_reduce, 1),
                shared=4 * threads_reduce,
            )

            correl[xiy, xix + disp_max + 1] = g_sumdata.get()

    for xiy in range(disp_max):
        dispy = xiy + 1
        nymax = ny1 - max(ny0 // 2 + ny1 // 2 + dispy - ny0, 0)
        ny1dep = 0
        ny0dep = ny0 // 2 - ny1 // 2 + dispy
        for xix in range(disp_max + 1):
            dispx = -disp_max + xix
            nxmax = nx1 + min(nx0 // 2 - nx1 // 2 + dispx, 0)
            nx1dep = -min(nx0 // 2 - nx1 // 2 + dispx, 0)
            nx0dep = max(0, nx0 // 2 - nx1 // 2 + dispx)

            threads_reduce = maxthreads
            blocks_reduce = nx0 * ny0
            if blocks_reduce < maxthreads:
                threads_reduce = nextpow2(blocks_reduce)
            blocks_reduce = (blocks_reduce + threads_reduce - 1) // threads_reduce
            correlate_cuda(
                g_im0data.gpudata,
                g_im1data.gpudata,
                g_odata.gpudata,
                np.int32(nx0),
                np.int32(nx1),
                np.int32(ny0),
                np.int32(dispx),
                np.int32(dispy),
                np.int32(nymax),
                np.int32(ny1dep),
                np.int32(ny0dep),
                np.int32(nx0dep),
                np.int32(nxmax),
                np.int32(nx1dep),
                block=(threads_per_block, threads_per_block, 1),
                grid=(blocks_per_tile, blocks_per_tile),
            )
            while blocks_reduce != 1:
                reduction_cuda(
                    g_odata.gpudata,
                    g_odata.gpudata,
                    block=(threads_reduce, 1, 1),
                    grid=(blocks_reduce, 1),
                    shared=4 * threads_reduce,
                )
                if blocks_reduce < maxthreads:
                    threads_reduce = nextpow2(blocks_reduce)
                blocks_reduce = (
                    blocks_reduce + threads_reduce - 1
                ) // threads_reduce
            reduction_cuda(
                g_odata.gpudata,
                g_sumdata.gpudata,
                block=(threads_reduce, 1, 1),
                grid=(blocks_reduce, 1),
                shared=4 * threads_reduce,
            )

            correl[xiy + disp_max + 1, xix] = g_sumdata.get()
        for xix in range(disp_max):
            dispx = xix + 1
            nxmax = nx1 - max(nx0 // 2 + nx1 // 2 + dispx - nx0, 0)
            nx1dep = 0
            nx0dep = nx0 // 2 - nx1 // 2 + dispx

            threads_reduce = maxthreads
            blocks_reduce = nx0 * ny0
            if blocks_reduce < maxthreads:
                threads_reduce = nextpow2(blocks_reduce)
            blocks_reduce = (blocks_reduce + threads_reduce - 1) // threads_reduce
            correlate_cuda(
                g_im0data.gpudata,
                g_im1data.gpudata,
                g_odata.gpudata,
                np.int32(nx0),
                np.int32(nx1),
                np.int32(ny0),
                np.int32(dispx),
                np.int32(dispy),
                np.int32(nymax),
                np.int32(ny1dep),
                np.int32(ny0dep),
                np.int32(nx0dep),
                np.int32(nxmax),
                np.int32(nx1dep),
                block=(threads_per_block, threads_per_block, 1),
                grid=(blocks_per_tile, blocks_per_tile),
            )
            while blocks_reduce != 1:
                reduction_cuda(
                    g_odata.gpudata,
                    g_odata.gpudata,
                    block=(threads_reduce, 1, 1),
                    grid=(blocks_reduce, 1),
                    shared=4 * threads_reduce,
                )
                if blocks_reduce < maxthreads:
                    threads_reduce = nextpow2(blocks_reduce)
                blocks_reduce = (
                    blocks_reduce + threads_reduce - 1
                ) // threads_reduce
            reduction_cuda(
                g_odata.gpudata,
                g_sumdata.gpudata,
                block=(threads_reduce, 1, 1),
                grid=(blocks_reduce, 1),
                shared=4 * threads_reduce,
            )

            correl[xiy + disp_max + 1, xix + disp_max + 1] = g_sumdata.get()

    return correl, norm



import os
import unittest

import h5py
import matplotlib.pyplot as plt

from fluidimage import get_path_image_samples
from fluidimage.calibration import CalibDirect, DirectStereoReconstruction
from fluidimage.calibration.util import get_plane_equation

pathbase = get_path_image_samples() / "4th_PIV-Challenge_Case_E"


def get_piv_field(path):

    try:
        with h5py.File(path, "r") as file:
            keyspiv = [key for key in file.keys() if key.startswith("piv")]
            keyspiv.sort()
            key = keyspiv[-1]
            X = file[key]["xs"][...]
            Y = file[key]["ys"][...]
            dx = file[key]["deltaxs_final"][...]
            dy = file[key]["deltays_final"][...]
    except Exception:
        print(path)
        raise

    return X, Y, dx, dy


class TestCalib(unittest.TestCase):
    """Test fluidimage.calibration DirectStereoReconstruction, CalibDirect."""

    def test(self):

        path_cam1 = pathbase / "E_Calibration_Images" / "Camera_01"
        path_cam3 = pathbase / "E_Calibration_Images" / "Camera_03"

        path_calib1 = path_cam1 / "calib1.npy"
        path_calib3 = path_cam3 / "calib3.npy"

        nb_pixelx, nb_pixely = 1024, 1024
        nbline_x, nbline_y = 32, 32

        calib = CalibDirect(path_cam1 / "img*", (nb_pixelx, nb_pixely))
        calib.compute_interpolents()
        calib.compute_interpolents_pixel2line(nbline_x, nbline_y, test=False)

        calib.check_interp_lines(4)
        plt.close("all")
        calib.check_interp_lines_coeffs(2)
        plt.close("all")
        calib.check_interp_levels(2)
        plt.close("all")

        calib3 = CalibDirect(
            os.path.join(path_cam3, "img*"), (nb_pixelx, nb_pixely)
        )

        calib3.compute_interpolents()
        calib3.compute_interpolents_pixel2line(nbline_x, nbline_y, test=False)

        calib.save(path_calib1)
        calib3.save(path_calib3)

        postfix = ".piv"
        name = "piv_00001-00002.h5"
        path_im = pathbase / "E_Particle_Images"

        path_piv1 = path_im / ("Camera_01" + postfix) / name
        path_piv3 = path_im / ("Camera_03" + postfix) / name

        z0 = 0
        alpha = 0
        beta = 0
        a, b, c, d = get_plane_equation(z0, alpha, beta)

        Xl, Yl, dxl, dyl = get_piv_field(path_piv1)
        Xr, Yr, dxr, dyr = get_piv_field(path_piv3)

        stereo = DirectStereoReconstruction(path_calib1, path_calib3)
        X0, X1, d0cam, d1cam = stereo.project2cam(
            Xl, Yl, dxl, dyl, Xr, Yr, dxr, dyr, a, b, c, d, check=False
        )
        X, Y, Z = stereo.find_common_grid(X0, X1, a, b, c, d)

        dx, dy, dz, erx, ery, erz = stereo.reconstruction(
            X0, X1, d0cam, d1cam, a, b, c, d, X, Y, check=False
        )


# dt = 0.001
# dx, dy, dz = dx/dt, dy/dt, dz/dt
# erx, ery, erz = erx/dt, ery/dt, erz/dt

if __name__ == "__main__":
    unittest.main()



import unittest
from pathlib import Path
from shutil import rmtree
from time import sleep

from fluidimage import get_path_image_samples
from fluidimage.topologies.piv import TopologyPIV

path_image_samples = get_path_image_samples()


class TestPivNew(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.path_Oseen = path_image_samples / "Oseen/Images/Oseen*"
        cls.path_Jet = path_image_samples / "Jet/Images/c06*"
        cls.postfix = "test_piv"

    @classmethod
    def tearDownClass(cls):
        paths = (cls.path_Oseen, cls.path_Jet)
        for path in paths:
            path_out = Path(str(path.parent) + "." + cls.postfix)
            if path_out.exists():
                rmtree(path_out)

    def test_piv_new(self):
        params = TopologyPIV.create_default_params()

        params.series.path = str(self.path_Oseen)
        params.series.ind_start = 1
        params.series.ind_step = 1

        params.piv0.shape_crop_im0 = 32
        params.multipass.number = 2
        params.multipass.use_tps = True

        # params.saving.how has to be equal to 'complete' for idempotent jobs
        # (on clusters)
        params.saving.how = "recompute"
        params.saving.postfix = self.postfix

        topology = TopologyPIV(params, logging_level="info")

        topology.make_code_graphviz(topology.path_dir_result / "topo.dot")
        topology.compute()

    def test_piv_new_multiproc(self):
        params = TopologyPIV.create_default_params()

        params.series.path = str(self.path_Jet)
        params.series.ind_start = 60
        params.series.ind_step = 1
        params.series.strcouple = "i, 0:2"

        params.piv0.shape_crop_im0 = 128
        params.multipass.number = 2
        params.multipass.use_tps = True

        params.saving.how = "recompute"
        params.saving.postfix = self.postfix

        topology = TopologyPIV(params, logging_level="info")
        topology.compute()

        topology = TopologyPIV(params, logging_level="info")
        topology.compute(nb_max_workers=2)

        # remove one file to test params.saving.how = "complete"
        path_files = list(Path(topology.path_dir_result).glob("piv*"))

        if not path_files:
            sleep(0.2)
            path_files = list(Path(topology.path_dir_result).glob("piv*"))

        path_files[0].unlink()

        params.saving.how = "complete"
        topology = TopologyPIV(params, logging_level="debug")
        topology.compute("exec_sequential")

        assert len(topology.results) == 1


if __name__ == "__main__":
    unittest.main()



import unittest
from pathlib import Path
from shutil import rmtree

from fluiddyn.io.image import imread, imsave
from fluidimage import get_path_image_samples
from fluidimage.topologies.preproc import TopologyPreproc


class TestPreprocTemporal(unittest.TestCase):
    name = "Jet"

    @classmethod
    def setUpClass(cls):
        path_in = get_path_image_samples() / cls.name / "Images"

        cls._work_dir = Path("test_topo_preproc_" + cls.name) / "Images"
        cls._work_dir.mkdir(parents=True, exist_ok=True)

        paths = path_in.glob("*")

        for path in paths:
            name = path.name
            im = imread(path)
            im = im[::6, ::6]
            imsave(str(cls._work_dir / name), im, as_int=True)

    @classmethod
    def tearDownClass(cls):
        rmtree(cls._work_dir.parent, ignore_errors=True)

    def test_preproc(self):
        """Test preproc subpackage on image sample Jet with two indices."""
        params = TopologyPreproc.create_default_params()

        params.preproc.series.path = self._work_dir
        params.preproc.series.strcouple = "i:i+2,1"
        params.preproc.series.ind_start = 60

        for tool in params.preproc.tools.available_tools:
            if "temporal" in tool:
                tool = params.preproc.tools.__getitem__(tool)
                tool.enable = True

        params.preproc.saving.how = "recompute"
        params.preproc.saving.postfix = "preproc_test"

        topology = TopologyPreproc(params, logging_level="debug")
        topology.compute("exec_async_sequential")
        assert len(topology.results) == 1

        params.preproc.saving.how = "complete"
        topology = TopologyPreproc(params, logging_level="debug")
        topology.compute("exec_async_sequential")
        assert len(topology.results) == 0


if __name__ == "__main__":
    unittest.main()



"""Preprocess toolbox (:mod:`fluidimage.pre_proc._toolbox_cv`)
==============================================================
A toolbox for preprocessing images. Based on OpenCV library.
cf. http://www.opencv.org.cn/opencvdoc/2.3.2/html/modules/imgproc/doc/filtering.html

.. currentmodule:: fluidimage.pre_proc._toolbox_cv

Provides:

   :members:

"""

import numpy as np

from .io import iterate_multiple_imgs
from .._opencv import cv2

__all__ = [
    "sliding_median",
    "sliding_minima",
    "global_threshold",
    "adaptive_threshold",
]


# ----------------------------------------------------
#   SPATIAL FILTERS
# ----------------------------------------------------


@iterate_multiple_imgs
def sliding_median(img=None, weight=1.0, window_size=3):
    """
    Subtracts the median calculated within a sliding window from the centre of
    the window.

    Parameters
    ----------
    img : array_like
        Single image as numpy array or multiple images as array-like object
    weight : scalar
        Fraction of median to be subtracted from each pixel.
        Value of `weight` should be in the interval (0.0,1.0).
    window_size : int
        Sets the size of the sliding window.

    """
    try:
        img_out = img - weight * cv2.medianBlur(
            img.astype(np.uint8), int(window_size)
        )
    except:
        print(
            "Check img dtype={}, shape={}, and window_size={}".format(
                img.dtype, img.shape, window_size
            )
        )
        raise

    return img_out


@iterate_multiple_imgs
def sliding_minima(
    img=None, weight=1.0, window_size=3, boundary_condition="reflect"
):
    """
    Subtracts the minimum calculated within a sliding window from the centre of
    the window.

    Parameters
    ----------
    img : array_like
        Single image as numpy array or multiple images as array-like object
    weight : scalar
        Fraction of minima to be subtracted from each pixel.
        Value of `weight` should be in the interval (0.0,1.0).
    window_size : scalar
        Sets the size of the sliding window.

    boundary_condition : {'reflect', 'default', 'constant', 'wrap',
                          'transparent', 'replicate'}

        Mode of handling array borders.

    """
    kernel = np.ones((window_size, window_size), np.uint8)
    border = getattr(cv2, f"BORDER_{boundary_condition.upper()}")
    img_out = img - weight * cv2.erode(
        img.astype(np.uint8), kernel=kernel, borderType=border
    )
    return img_out


@iterate_multiple_imgs
def global_threshold(img=None, minima=0.0, maxima=65535.0):
    """
    Trims pixel intensities which are outside the interval (minima, maxima).

    Parameters
    ----------
    img : array_like
        Single image as numpy array or multiple images as array-like object

    minima, maxima : float
        Sets the threshold

    """
    MAX_OPER = cv2.THRESH_TRUNC
    MIN_OPER = cv2.THRESH_TOZERO
    img_out = cv2.threshold(img, thresh=maxima, maxval=maxima, type=MAX_OPER)
    img_out = cv2.threshold(img_out, thresh=minima, maxval=maxima, type=MIN_OPER)
    return img_out


@iterate_multiple_imgs
def adaptive_threshold(img=None, window_size=5, offset=0):
    """
    Adaptive threshold transforms a grayscale image to a binary image.
    Useful in identifying particles.

    Parameters
    ----------
    img : array_like
        Single image as numpy array or multiple images as array-like object
    window_size : scalar
        Sets the size of the pixel neighbourhood to calculate threshold.
    offset : scalar
        Constant to be subtracted from the mean.

    """
    adaptiveMethod = cv2.ADAPTIVE_THRESH_GAUSSIAN_C
    thresholdType = cv2.THRESH_BINARY  # cv2.THRESH_BINARY_INV
    img_out = cv2.adaptiveThreshold(
        img, img.max(), adaptiveMethod, thresholdType, window_size, offset
    )
    return img_out



import matplotlib.pyplot as plt
import numpy as np
from numpy import pi

from fluidimage.calcul.interpolate import thin_plate_spline_subdom


def myplot(i, x, y, U, title=None):
    plt.figure(i)
    ax = plt.gca()

    ax.scatter(x, y, c=U, vmin=-1, vmax=1)

    if title is not None:
        ax.set_title(title)


# set of random x coordinates from 0 to 2pi
x = 2 * np.pi * np.random.rand(100)
y = 2 * np.pi * np.random.rand(100)

U = np.exp(-((x - pi) ** 2 + (y - pi) ** 2))
V = np.exp(-((x - pi / 2) ** 2 + (y - pi / 2) ** 2))

myplot(0, x, y, U, "input U")
myplot(1, x, y, V, "input V")

# calculate tps coeff
centers = np.vstack([x, y])
smoothing_coef = 0
subdom_size = 20

tps = thin_plate_spline_subdom.ThinPlateSplineSubdom(
    centers, subdom_size, smoothing_coef, threshold=1, pourc_buffer_area=0.5
)

U_smooth, U_tps = tps.compute_tps_coeff_subdom(U)
V_smooth, V_tps = tps.compute_tps_coeff_subdom(V)

# interpolation grid
xI = yI = np.arange(0, 2 * pi, 0.1)
XI, YI = np.meshgrid(xI, yI)
XI = XI.ravel()
YI = YI.ravel()

new_positions = np.vstack([XI, YI])

tps.init_with_new_positions(new_positions)

U_eval = tps.compute_eval(U_tps)
V_eval = tps.compute_eval(V_tps)

myplot(2, XI, YI, U_eval, "U_eval")
myplot(3, XI, YI, V_eval, "V_eval")

plt.show()



"""Simple viewer adapted for fluid images (:mod:`fluidimage.gui.imviewer`)
==========================================================================

Coded with matplotlib GUI!

.. autoclass:: ImageViewer
   :members:
   :private-members:

"""

import argparse
import os
from glob import glob

import matplotlib.pyplot as plt
from matplotlib.widgets import TextBox, Button

from fluiddyn.io.image import imread
from fluiddyn.util import time_as_str
from fluiddyn.util.serieofarrays import SerieOfArraysFromFiles

extensions = ["png", "tif", "tiff", "jpg", "jpeg", "bmp", "cine"]
extensions = ["." + ext for ext in extensions]

debug = False


def _print_debug(*args):
    if debug:
        print(*args)


def check_image(path):
    return any([path.endswith(ext) for ext in extensions])


size_button = 0.06
sep_button = 0.02
x0 = 0.3

name_buttons = ["-n", "-1", "+1", "+n"]

x_buttons = [
    x0 + i * (size_button + sep_button) for i in range(len(name_buttons))
]


def parse_args():
    parser = argparse.ArgumentParser(
        description=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument(
        "path",
        help="Path file or directory.",
        type=str,
        nargs="?",
        default=os.getcwd(),
    )
    parser.add_argument(
        "-cm", "--colormap", help="colormap", type=str, default=""
    )
    parser.add_argument("-v", "--verbose", help="verbose mode", action="count")

    return parser.parse_args()


class ImageViewer:
    """Simple Image viewer."""

    def __init__(self, args):

        path_in = args.path
        if os.path.isdir(path_in):
            self.path_files = glob(os.path.join(path_in, "*"))
            self.path_files = [
                path for path in self.path_files if check_image(path)
            ]
            self.path_files.sort()
            ifile = 0
        else:
            path_file = glob(path_in)[0]
            self.path_files = glob(os.path.join(os.path.split(path_file)[0], "*"))
            self.path_files = [
                path for path in self.path_files if check_image(path)
            ]
            self.path_files.sort()
            ifile = self.path_files.index(path_file)

        if len(self.path_files) == 0:
            raise ValueError("No image files detected.")

        if len(self.path_files) == 1 and self.path_files[0].endswith(".cine"):
            serie = SerieOfArraysFromFiles(self.path_files[0])
            self.path_files = serie.get_path_arrays()

        path_dir = os.path.split(self.path_files[0])[0]
        self.nb_images = len(self.path_files)
        print(f"Will use {self.nb_images} images in the dir {path_dir}")

        self._buttons = {}
        self._textboxes = {}

        fig = self.fig = plt.figure()
        fig.canvas.manager.set_window_title(
            path_dir + " (" + time_as_str()[-8:].replace("-", ":") + ")"
        )
        self.ax = fig.add_axes([0.07, 0.15, 0.7, 0.78])

        self.maps = {}
        try:
            self.cmap = plt.cm.viridis
        except AttributeError:
            self.cmap = plt.cm.jet

        self.ifile = ifile
        self._last_was_increase = False

        im = imread(self.path_files[ifile])
        self.clim = [0, 0.99 * im.max()]
        self._updating_clim = False

        self.loadim(ifile, im)
        name_file = self.get_namefile()
        self.ax.set_title(name_file)

        self._image_changing = False

        function_buttons = [None] * len(name_buttons)
        function_buttons[0] = self._decrease_ifile_n
        function_buttons[1] = self._decrease_ifile
        function_buttons[2] = self._increase_ifile
        function_buttons[3] = self._increase_ifile_n

        y = size_button / 3.0
        for i, x in enumerate(x_buttons):
            name = name_buttons[i]
            func = function_buttons[i]
            self._create_button(fig, [x, y, size_button, size_button], name, func)

        self._n = 1

        self._create_text(
            fig,
            [0.1, y, 2 * size_button, size_button],
            "n = ",
            self._submit_n,
            "1",
        )

        self._create_text(
            fig,
            [0.87, 0.92, 1.5 * size_button, size_button],
            "cmax = ",
            self._change_cmax,
            "{:.2f}".format(self.clim[1]),
        )

        self._create_text(
            fig,
            [0.87, 0.1, 1.5 * size_button, size_button],
            "cmin = ",
            self._change_cmin,
            "{:.2f}".format(self.clim[0]),
        )

        self._create_button(
            fig, [0.65, 0.945, 1.2 * size_button, 0.045], "reload", self.reloadim
        )

        self._create_button(
            fig, [0.85, y, size_button, size_button], "auto", self.set_autoclim
        )

        cax = fig.add_axes([0.83, 0.2, 0.07, 0.7])
        self.cbar = fig.colorbar(self.mappable, cax=cax)

        fig.canvas.mpl_connect("key_press_event", self.onclick)
        print("press alt+h for help")

        plt.show()

    def set_autoclim(self, event):
        _print_debug("set_autoclim")
        im = imread(self.path_files[self.ifile])
        self.clim = [im.min(), 0.99 * im.max()]
        self._update_clim()

    def reloadim(self, event):
        self.loadim(self.ifile)

    def loadim(self, ifile, im=None):
        _print_debug("loadim", ifile, im)
        if im is None:
            im = imread(self.path_files[ifile])
            _print_debug(self.path_files[ifile])
            _print_debug(im)
        self.mappable = self.ax.imshow(
            im,
            interpolation="nearest",
            cmap=self.cmap,
            origin="upper",
            extent=[0, im.shape[1], im.shape[0], 0],
            vmin=self.clim[0],
            vmax=self.clim[1],
        )
        self.maps[ifile] = self.mappable

    def get_namefile(self):
        return os.path.split(self.path_files[self.ifile])[-1]

    def change_im(self):
        self.ifile = self.ifile % self.nb_images
        if self._image_changing:
            return

        self._image_changing = True
        ifile = self.ifile
        name_file = self.get_namefile()
        print("changing to file " + name_file, end="...")
        _print_debug("")
        map_old = self.mappable

        if ifile in self.maps:
            self.mappable = self.maps[ifile]
            self.mappable.set_clim(self.clim)
            self.mappable.set_visible(True)
        else:
            self.loadim(ifile)

        map_old.set_visible(False)
        self.ax.set_title(name_file)
        self.fig.canvas.draw()
        _print_debug("")
        print("\rchanged to file " + name_file + " " * 20)
        self._image_changing = False

    def _create_button(self, fig, rect, text, func):
        ax = fig.add_axes(rect)
        button = Button(ax, text)
        button.on_clicked(func)
        self._buttons[text] = button
        return button

    def _create_text(self, fig, rect, name, func, initial):
        ax = fig.add_axes(rect)
        textbox = TextBox(ax, name, initial=initial)
        textbox.on_submit(func)
        self._textboxes[name] = textbox
        return textbox

    def _switch(self):
        if self._last_was_increase:
            self._decrease_ifile_n()
        else:
            self._increase_ifile_n()
        self._last_was_increase = not self._last_was_increase

    def _increase_ifile(self, event=None):
        self.ifile += 1
        self.change_im()

    def _decrease_ifile(self, event=None):
        self.ifile -= 1
        self.change_im()

    def _increase_ifile_n(self, event=None):
        self.ifile += self._n
        self.change_im()

    def _decrease_ifile_n(self, event=None):
        self.ifile -= self._n
        self.change_im()

    def _submit_n(self, text):
        self._n = eval(text, {}, {})
        print("n = ", self._n)

    def _change_cmax(self, text):
        _print_debug("_change_cmax")
        cmax = eval(text, {}, {})
        if cmax == self.clim[1]:
            return

        self.clim[1] = cmax
        self._update_clim()

    def _change_cmin(self, text):
        _print_debug("_change_cmin")
        cmin = eval(text, {}, {})

        if cmin == self.clim[0]:
            return

        self.clim[0] = cmin
        self._update_clim()

    def _update_clim(self):
        if self._updating_clim:
            return

        self._updating_clim = True
        _print_debug("_update_clim:", self.clim)
        self.mappable.set_clim(self.clim)
        self.fig.canvas.draw()
        # self.cbar.set_clim(self.clim)
        self.fig.draw_without_rendering()

        txt_cmax = self._textboxes["cmax = "]
        txt_cmin = self._textboxes["cmin = "]
        txt_cmin.set_val("{:.2f}".format(self.clim[0]))
        txt_cmax.set_val("{:.2f}".format(self.clim[1]))
        _print_debug("_update_clim end")
        self._updating_clim = False

    def onclick(self, event):
        if event.key == "alt+h":
            print("\nalt+s\t\t\t switch between images\n")

        if event.inaxes != self.ax:
            return

        if event.key == "alt+s":
            self._switch()


# if debug and __name__ == '__main__':
#     args = parse_args()
#     self = ImageViewer(args)



"""MLOS (Multiplicative Line of Sight)
======================================

Reference::

    C. Atkinson and J. Soria, “An efficient simultaneous reconstruction
    technique for tomographic particle image velocimetry,” Exp Fluids, vol. 47,
    no. 4–5, p. 553, Oct. 2009.


.. autoclass:: TomoMLOSBase
   :members:
   :private-members:

.. autoclass:: TomoMLOSRbf
   :members:
   :private-members:

.. autoclass:: TomoMLOSNeighbour
   :members:
   :private-members:

.. autoclass:: TomoMLOSCV
   :members:
   :private-members:

"""

import os
from pathlib import Path

import dask.array as da
import matplotlib.pyplot as plt
import numpy as np
from dask.diagnostics import ProgressBar
from psutil import virtual_memory
from scipy import interpolate, sparse

from fluidimage._opencv import cv2
from fluidimage.calibration.calib_cv import CalibCV
from fluidimage.data_objects.tomo import ArrayTomoCV
from fluidimage.util import imread


class TomoMLOSBase:
    """MLOS can be summarized in the following steps:

    1. Project the world coordinates to pixel coordinates.

    2. Interpolate intensity of the neighbouring pixels.

    3. Project back the interpolated intesities onto world coordinates and
       apply them multiplicatively.

    """

    def __init__(self, cls_calib, cls_array, *cams, **kwargs):
        self.array = cls_array(params=None, **kwargs)
        self.cams = []
        for cam in cams:
            cam_name = Path(cam).name.split(".")[0]
            self.cams.append(cam_name)
            setattr(self, cam_name, cls_calib(cam))

    def reconstruct(
        self,
        pix: dict,
        image: np.ndarray,
        threshold: float,
        chunks: tuple,
        save: bool,
    ):
        """Performs MLOS reconstruction parallely using Dask. The
        reconstruction is done in memory when `save=False` and in the
        filesystem when `save=True`.

        """
        self.array.image_path = image
        interp = self.get_interpolator(image, threshold)
        with ProgressBar():
            x_pix = da.from_array(pix["x"], chunks=chunks)
            y_pix = da.from_array(pix["y"], chunks=chunks)
            i_vox = da.map_blocks(interp, x_pix, y_pix)
            i_vox = i_vox ** (1.0 / len(self.cams))
            if save:
                if os.path.exists(self.array.h5file_path):
                    # Create a temporary copy to read from
                    file, dset, h5file_tmp = self.array.load_dataset(copy=True)
                    self.array.I = da.from_array(dset, chunks=chunks)
                    self.array.I *= i_vox
                    self.array.save()
                    file.close()
                    os.remove(h5file_tmp)
                else:
                    self.array.I = i_vox
                    self.array.save()
            else:
                self.array.I *= i_vox.compute()

    def verify_projection(self, cam="cam0", skip=1):
        """Graphically verify the projection performed by `phys2pix` method."""
        fig, axes = plt.subplots(1, 3, figsize=(15, 4))
        cmap = plt.get_cmap("gray")

        pix = self.phys2pix(cam)

        titles = [r + r"$_{world}$" for r in ("x", "y", "z")]
        for i, axt in enumerate(zip(axes, titles)):
            ax, title = axt
            try:
                r_world = self.array.grid[::skip, i]
            except AttributeError:
                key = ("X", "Y", "Z")[i]
                r_world = getattr(self.array, key).flatten()[::skip]

            color = cmap(r_world / r_world.max())
            ax.scatter(
                pix["x"].flatten()[::skip], pix["y"].flatten()[::skip], c=color
            )
            ax.set_title(title)
            # ax.pcolor(pix["x"], pix["y"], r_world)
            ax.set_xlabel(r"$x_{pix}$")
            ax.set_ylabel(r"$y_{pix}$")


class TomoMLOSRbf(TomoMLOSBase):
    """Interpolation is calculated using a radial basis function (RBF) interpolation
    with a gaussian kernel.

    """

    def get_interpolator(self, image, threshold):
        im = imread(image)

        if threshold is not None:
            im[im < threshold] = 0.0
        im_sparse = sparse.coo_matrix(im)

        # Interpolation function
        return interpolate.Rbf(
            im_sparse.col, im_sparse.row, im_sparse.data, function="gaussian"
        )


class TomoMLOSNeighbour(TomoMLOSBase):
    """Interpolation is calculated using a nearest neighbour interpolation."""

    def get_interpolator(self, image, threshold):
        im = imread(image)

        if threshold is not None:
            im[im < threshold] = 0.0
        im_sparse = sparse.coo_matrix(im)

        # Interpolation function
        return interpolate.NearestNDInterpolator(
            (im_sparse.col, im_sparse.row), im_sparse.data
        )


class TomoMLOSCV(TomoMLOSNeighbour):
    def __init__(self, *cams, **kwargs):
        super().__init__(CalibCV, ArrayTomoCV, *cams, **kwargs)

    def phys2pix(self, cam_name: str):
        """Tranform the 'physical' world coordinates to 'pixel' coordinates."""
        cam = getattr(self, cam_name)
        grid3d = self.array.grid
        pix = np.empty((len(grid3d), 1, 2))

        for i, z in enumerate(self.array.zs):
            grid_zslice = grid3d[grid3d[:, 2] == z]
            n = len(grid_zslice)
            istart = i * n
            iend = (i + 1) * n
            rotation = cam.get_rotation(z)
            translate = cam.get_translate(z)
            pix[istart:iend, ...], jac = cv2.projectPoints(
                grid_zslice,
                rotation,
                translate,
                cam.params.cam_mtx,
                cam.params.kc,
            )
        return {"x": pix[:, 0, 0], "y": pix[:, 0, 1]}

    def reconstruct(
        self,
        pix: dict,
        image: np.ndarray,
        threshold=None,
        chunks=None,
        save=False,
    ):
        """Estimate the maximum size of chunk which can fit in the memory and
        execute the parent method.

        """
        if chunks is None:
            chunks = len(pix["x"]) // os.cpu_count()
            nmax = int(_estimate_max_array_size(self.array.dtype)) // 6
            if chunks > nmax:
                chunks = nmax

        super().reconstruct(pix, image, threshold, chunks, save)


def _estimate_max_array_size(dtype=np.float64):
    mem = virtual_memory()
    nbytes = np.array([0], dtype=dtype).nbytes
    return mem.available // nbytes


if __name__ == "__main__":
    import shutil
    from fluidimage import get_path_image_samples

    path_image_samples = get_path_image_samples()

    path = path_image_samples / "TomoPIV" / "calibration"
    cameras = [str(path / f"cam{i}.h5") for i in range(4)]
    tomo = TomoMLOSCV(
        *cameras,
        xlims=(-10, 10),
        ylims=(-10, 10),
        zlims=(-5, 5),
        nb_voxels=(20, 20, 10),
    )
    tomo.verify_projection()

    particle_images = path_image_samples / "TomoPIV" / "particle"
    output_dir = path_image_samples / "TomoPIV" / "tomo"
    if output_dir.exists():
        shutil.rmtree(output_dir)

    for cam in tomo.cams:
        print(f"Projecting {cam}...")
        pix = tomo.phys2pix(cam)
        i0 = 1
        for i1 in ["a", "b"]:
            image = str(particle_images / f"{cam}.pre" / f"im{i0:05.0f}{i1}.tif")
            tomo.array.init_paths(image)
            print(f"MLOS of {cam} on {image}: reconstructing...")
            tomo.reconstruct(pix, image, threshold=None, save=True)



"""Display pairs of images (:mod:`fluidimage.data_objects.display_pre`)
=======================================================================

.. autoclass:: DisplayPreProc
   :members:
   :private-members:

"""

import matplotlib.pyplot as plt
import numpy as np


class DisplayPreProc:
    """Display pairs of images"""

    def __init__(self, im0, im1, im0p, im1p, pourcent_histo=99, hist=False):

        fig = plt.figure()
        fig.event_handler = self

        ax1 = plt.subplot(121)
        ax2 = plt.subplot(122)

        self.fig = fig
        self.ax1 = ax1
        self.ax2 = ax2

        p0 = np.percentile(
            np.reshape(im0, (1, np.product(im0.shape))).transpose(),
            pourcent_histo,
        )
        p1 = np.percentile(
            np.reshape(im1, (1, np.product(im1.shape))).transpose(),
            pourcent_histo,
        )
        p0p = np.percentile(
            np.reshape(im0p, (1, np.product(im0p.shape))).transpose(),
            pourcent_histo,
        )
        p1p = np.percentile(
            np.reshape(im1p, (1, np.product(im1p.shape))).transpose(),
            pourcent_histo,
        )
        im0[im0 > p0] = p0
        im1[im1 > p1] = p1
        im0p[im0p > p0p] = p0p
        im1p[im1p > p1p] = p1p

        self.image0 = ax1.imshow(
            im0,
            interpolation="nearest",
            cmap=plt.cm.gray,
            origin="upper",
            extent=[0, im0.shape[1], im0.shape[0], 0],
        )

        self.image1 = ax1.imshow(
            im1,
            interpolation="nearest",
            cmap=plt.cm.gray,
            origin="upper",
            extent=[0, im0.shape[1], im0.shape[0], 0],
        )
        self.image1.set_visible(False)

        self.image0p = ax2.imshow(
            im0p,
            interpolation="nearest",
            cmap=plt.cm.gray,
            origin="upper",
            extent=[0, im0p.shape[1], im0p.shape[0], 0],
        )

        self.image1p = ax2.imshow(
            im1p,
            interpolation="nearest",
            cmap=plt.cm.gray,
            origin="upper",
            extent=[0, im0p.shape[1], im0p.shape[0], 0],
        )
        self.image1p.set_visible(False)

        if hist:
            fig2 = plt.figure()
            ax3 = plt.subplot(121)
            ax4 = plt.subplot(122)
            hist0 = np.histogram(
                np.reshape(im0, (1, np.product(im0.shape))).transpose(), bins="fd"
            )
            hist1 = np.histogram(
                np.reshape(im1, (1, np.product(im1.shape))).transpose(), bins="fd"
            )
            hist0p = np.histogram(
                np.reshape(im0p, (1, np.product(im0p.shape))).transpose(),
                bins="fd",
            )
            hist1p = np.histogram(
                np.reshape(im1p, (1, np.product(im1p.shape))).transpose(),
                bins="fd",
            )
            incr = 1
            ax3.plot(hist0[1][0:-1:incr], hist0[0][0::incr], "k+")
            ax3.plot(hist0p[1][0:-1:incr], hist0p[0][0::incr], "r+")
            ax4.plot(hist1[1][0:-1:incr], hist1[0][0::incr], "k+")
            ax4.plot(hist1p[1][0:-1:incr], hist1p[0][0::incr], "r+")

            ax3.set_xlim(-10, max([p0, p0p]))
            ax4.set_xlim(-10, max([p1, p1p]))
            ax3.set_ylim(0, max(hist0[0]))
            ax4.set_ylim(0, max(hist1[0]))
            fig2.show()

        (l,) = ax1.plot(0, 0, "oy")
        l.set_visible(False)

        ax1.set_title("im 0 (alt+s to switch)")

        ax1.set_xlim(0, im0.shape[1])
        ax1.set_ylim(im0.shape[0], 0)
        ax1.set_xlabel("pixels")
        ax1.set_ylabel("pixels")

        ax1.set_xlim(0, im0.shape[1])
        ax1.set_ylim(im0.shape[0], 0)

        ax1.set_xlabel("pixels")
        ax1.set_ylabel("pixels")

        (l,) = ax2.plot(0, 0, "oy")
        l.set_visible(False)

        ax2.set_title("im 0p (alt+s to switch)")

        ax2.set_xlim(0, im0p.shape[1])
        ax2.set_ylim(im0p.shape[0], 0)
        ax2.set_xlabel("pixels")
        ax2.set_ylabel("pixels")

        self.ind = 0
        fig.canvas.mpl_connect("key_press_event", self.onclick)
        print("press alt+h for help")

        plt.show()

    def onclick(self, event):
        if event.key == "alt+s":
            self.switch()

    def switch(self):
        self.image0.set_visible(not self.image0.get_visible())
        self.image1.set_visible(not self.image1.get_visible())

        self.ax1.set_title(
            "im {} (alt+s to switch)".format(int(self.image1.get_visible()))
        )

        self.fig.canvas.draw()

        self.image0p.set_visible(not self.image0p.get_visible())
        self.image1p.set_visible(not self.image1p.get_visible())

        self.ax2.set_title(
            "im {}p (alt+s to switch)".format(int(self.image1p.get_visible()))
        )
        self.fig.canvas.draw()



"""Executor async/await sequential (:mod:`fluidimage.executors.exec_async_sequential`)
======================================================================================

A executor using async for IO but launching CPU-bounded tasks sequentially.

.. autoclass:: ExecutorAsyncSequential
   :members:
   :private-members:

"""

import time

from fluidimage.util import cstring, log_memory_usage, logger

from .exec_async import ExecutorAsync


class ExecutorAsyncSequential(ExecutorAsync):
    """Async executor launching CPU-bounded tasks sequentially"""

    async def async_run_work_cpu(self, work):
        """Executes the work on an item (key, obj), and add the result on
        work.output_queue.

        Parameters
        ----------

        work :

          A work from the topology

        key : hashable

          The key of the dictionnary item to be process

        obj : object

          The value of the dictionnary item to be process

        """
        self.nb_working_workers_cpu += 1

        try:
            key, obj = work.input_queue.pop_first_item()
        except KeyError:
            self.nb_working_workers_cpu -= 1
            return

        if work.check_exception(key, obj):
            self.nb_working_workers_cpu -= 1
            return

        t_start = time.time()
        log_memory_usage(
            f"{time.time() - self.t_start:.2f} s. Launch work "
            + work.name_no_space
            + f" ({key}). mem usage"
        )
        # pylint: disable=W0703
        try:
            # here we do something very bad from the async point of view:
            # we launch a potentially long blocking function:
            ret = work.func_or_cls(obj)
        except Exception as error:
            self.log_exception(error, work.name_no_space, key)
            if self.stop_if_error:
                raise
            ret = error
        else:
            logger.info(
                f"work {work.name_no_space} ({key}) "
                f"done in {time.time() - t_start:.3f} s"
            )

        if work.output_queue is not None:
            work.output_queue[key] = ret
        self.nb_working_workers_cpu -= 1



"""Works Preprocess (:mod:`fluidimage.works.preproc`)
=====================================================

To preprocess series of images using topology.

Provides:

.. autoclass:: WorkPreproc
   :members:
   :private-members:

"""
import sys

import numpy as np

from fluiddyn.util.serieofarrays import SerieOfArraysFromFiles

from ..data_objects.preproc import ArraySerie, PreprocResults, get_ind_middle
from ..preproc.base import PreprocBase, _make_doc_with_filtered_params_doc
from ..util import print_memory_usage


class WorkPreproc(PreprocBase):
    """Work for preprocessing."""

    def calcul(self, serie):
        """Apply all enabled preprocessing tools on the series of arrays
        and returns the result as a data object.

        """
        if isinstance(serie, SerieOfArraysFromFiles):
            serie = ArraySerie(serie=serie)

        if not isinstance(serie, ArraySerie):
            raise ValueError("serie must be an instance of class ArraySerie")

        result = PreprocResults(self.params)
        images = np.array(serie.get_arrays())
        images = self.tools(images)
        serie._clear_data()
        dico = self._make_dict_to_save(serie, images)
        result.data.update(dico)
        print_memory_usage(
            "Memory usage after preprocessing {}/{} series".format(
                serie.ind_serie + 1, serie.nb_series
            )
        )
        return result

    def _make_dict_to_save(self, array_serie, images):
        name_files = array_serie.names
        nb_series = array_serie.nb_series
        ind_serie = array_serie.ind_serie

        ind_middle_start, ind_middle_end = get_ind_middle(array_serie.serie)
        if ind_serie == 0 and nb_series == 1:
            s = slice(None, None)
        elif ind_serie == 0:
            s = slice(0, ind_middle_end)
        elif ind_serie == nb_series - 1:
            s = slice(ind_middle_start, None)
        else:
            s = slice(ind_middle_start, ind_middle_end)

        return dict(zip(name_files[s], images[s]))

    def display(self, ind=0, hist=False):
        nb_images = 2
        name_files = self.serie_arrays.get_name_files()[ind : ind + nb_images]

        results_series = SerieOfArraysFromFiles(self.params.saving.path)
        results = {
            name: results_series.get_array_from_name(name) for name in name_files
        }

        return super().display(ind, hist, results)


Work = WorkPreproc

if "sphinx" in sys.modules:
    __doc__ += _make_doc_with_filtered_params_doc(Work)



"""Import OpenCV
================
Convenience module to import OpenCV and warn if the package was not found.
Hint: install it!

"""

from fluiddyn.util.opencv import cv2, error_import_cv2

if isinstance(error_import_cv2, ModuleNotFoundError):
    print(
        "Warning: OpenCV must be built and installed with python bindings "
        "to use this module. Install using either of the following\n"
        "    pip install opencv-python\n"
        "    conda install -c conda-forge opencv"
    )



"""Preprocess toolbox (:mod:`fluidimage.preproc.toolbox`)
=========================================================

A toolbox of filters which operate on a single image (numpy array).
cf. http://www.scipy-lectures.org/advanced/image_processing/

Provides:

.. autoclass:: PreprocToolsBase
   :members:

.. autoclass:: PreprocToolsPy
   :members:

.. autoclass:: PreprocToolsCV
   :members:

"""

import inspect

from ..util import logger


class PreprocToolsBase:
    """Base class for wrapping preprocessing functions into a class."""

    @classmethod
    def _get_backend(cls):
        raise NotImplementedError

    @classmethod
    def create_default_params(cls, params):
        """Create default parameters from the function argument list."""
        tools = cls._get_backend()
        available_tools = tools.__all__

        params.preproc._set_child("tools")
        params = params.preproc.tools
        params._set_attribs({"available_tools": tools.__all__, "sequence": None})

        for tool in available_tools:
            func = tools.__dict__[tool]

            # TODO: Replace with inspect.getfullargspec (Python >= 3).
            func_args = inspect.getcallargs(func)
            for arg in list(func_args.keys()):
                if arg in ["img"]:
                    # Remove arguments which are not parameters
                    del func_args[arg]

            func_args.update({"enable": False})

            # New parameter child for each tool and parameter attributes
            # from its function arguments and default values
            params._set_child(tool, attribs=func_args)

            # Adds docstring to the parameter
            try:
                func_doc = func.func_doc
            except AttributeError:
                func_doc = func.__doc__

            if func_doc is not None:
                enable_doc = (
                    "enable : bool\n" + "        Set as `True` to enable the tool"
                )
                params[tool]._set_doc(func_doc + enable_doc)

    @classmethod
    def _complete_class_with_tools(cls):
        """
        Dynamically add the global functions in this module as staticmethods of
        the present class.

        """
        tools = cls._get_backend()
        available_tools = tools.__all__

        for tool in available_tools:
            func = tools.__dict__[tool]

            # Add tools as `staticmethods` of the class
            setattr(cls, tool, func)

    def __init__(self, params):
        self.params = params.preproc.tools

    def __call__(self, img):
        """
        Apply all preprocessing tools for which `enable` is `True`.
        Return the preprocessed image (numpy array).

        Parameters
        ----------
        img : array_like
            Single image as numpy array or multiple images as array-like object

        """
        sequence = self.params.sequence
        if sequence is None:
            sequence = self.params.available_tools

        for tool in sequence:
            tool_params = self.params[tool]
            if tool_params.enable:
                logger.debug("Apply " + tool)
                kwargs = tool_params._make_dict_attribs()
                for k in list(kwargs.keys()):
                    if k == "enable":
                        kwargs.pop(k)

                cls = self.__class__
                img = cls.__dict__[tool](img, **kwargs)

        return img


class PreprocToolsPy(PreprocToolsBase):
    """Wrapper class for functions in _toolbox_py module."""

    @classmethod
    def _get_backend(cls):
        from . import _toolbox_py as tools

        return tools


PreprocToolsPy._complete_class_with_tools()


class PreprocToolsCV(PreprocToolsBase):
    """Wrapper class for functions in _toolbox_cv module."""

    @classmethod
    def _get_backend(cls):
        from . import _toolbox_cv as tools

        return tools


PreprocToolsCV._complete_class_with_tools()



""""Image to image" processing (:mod:`fluidimage.preproc.image2image`)
======================================================================

.. autofunction:: init_im2im_function

.. autofunction:: complete_im2im_params_with_default

.. autofunction:: apply_im2im_filter

.. autofunction:: im2im_func_example

.. autoclass:: Im2ImExample
   :members:

"""

import types

import numpy as np

from fluiddyn.util import import_class
from fluiddyn.util.serieofarrays import SerieOfArraysFromFiles

from ..data_objects.piv import ArrayCouple


def im2im_func_example(tuple_image_path):
    """Process one image

    This is just an example to show how to write functions which can be plugged
    to the class
    :class:`fluidimage.topologies.image2image.TopologyImage2Image`.

    """
    image, path = tuple_image_path
    # the processing can be adjusted depending on the value of the path.
    print("process file:\n" + path)
    image_out = np.round(image * (255 / image.max())).astype(np.uint8)
    return image_out, path


class Im2ImExample:
    """Process one image

    This is just an example to show how to write classes which can be plugged
    to the class
    :class:`fluidimage.topologies.image2image.TopologyImage2Image`.

    """

    def __init__(self, arg0, arg1):
        print("init with arguments:", arg0, arg1)
        self.arg0 = arg0
        self.arg1 = arg1

    # time consuming tasks can be done here

    def calcul(self, tuple_image_path):
        """Method processing one image"""
        print(
            "calcul with arguments (unused in the example):", self.arg0, self.arg1
        )
        return im2im_func_example(tuple_image_path)


def complete_im2im_params_with_default(params):
    """Complete params for image-to-image processing."""

    params._set_attrib("im2im", None)
    params._set_attrib("args_init", tuple())

    params._set_doc(
        """
im2im : str {None}

    Function or class to be used to process the images.

args_init : object {None}

    An argument given to the init function of the class used to process the
    images.

"""
    )


def init_im2im_function(im2im=None, args_init=()):
    """Initialize the filter function."""

    if isinstance(im2im, str):
        str_package, str_obj = im2im.rsplit(".", 1)
        im2im = import_class(str_package, str_obj)

    if isinstance(im2im, types.FunctionType):
        obj = im2im
        im2im_func = im2im
    elif isinstance(im2im, type):
        print("in init_im2im", args_init)
        obj = im2im(*args_init)
        im2im_func = obj.calcul

    return obj, im2im_func


def apply_im2im_filter(serie, im2im=None, args_init=()):
    """Apply an image-to-image filter to a serie of images.

    Parameters
    ----------

    serie : :class:`fluiddyn.util.serieofarrays.SerieOfArraysFromFiles`

    im2im : Optional[str]

    args_init : tuple

    """
    if im2im is None:
        return serie

    obj, im2im_func = init_im2im_function(im2im, args_init)

    if not isinstance(serie, SerieOfArraysFromFiles):
        raise NotImplementedError

    arrays = serie.get_arrays()
    paths = serie.get_path_arrays()

    return tuple(im2im_func(t) for t in zip(arrays, paths))



import copy

import numpy as np

from .util import make_params_calibration


class Calibration:
    def __init__(self, path_file):
        self.path_file = path_file
        self.params = make_params_calibration(path_file)

    def pix2phys_UV(self, X, Y, dx, dy, index_level, nbypix, angle=True):
        """Apply Tsai Calibration to the field

        Notes
        -----

        The displacement is 3 component BUT it corresponds to the real 3
        component displacement projected on the corresponding plane indexed by
        index_level.

        """

        Xphys, Yphys, Zphys = self.pix2phys(
            X, Y, index_level=index_level, nbypix=nbypix, angle=angle
        )
        dxb, dyb, dzb = self.pix2phys(
            X + dx / 2.0, Y + dy / 2.0, index_level, nbypix=nbypix, angle=angle
        )
        dxa, dya, dza = self.pix2phys(
            X - dx / 2.0, Y - dy / 2.0, index_level, nbypix=nbypix, angle=angle
        )
        dxphys = dxb - dxa
        dyphys = dyb - dya
        dzphys = dzb - dza
        return Xphys, Yphys, Zphys, dxphys, dyphys, dzphys

    def pix2phys(self, X, Y, index_level, nbypix, angle=True):
        params = copy.deepcopy(self.params)
        # difference of convention with calibration done with uvmat!
        Y = nbypix - Y
        # determine position of Z0
        testangle = 0
        if (
            hasattr(params.slices, "slice_angle")
            and np.any(
                params.slices.slice_angle[index_level] != np.asarray([0, 0, 0])
            )
            and angle
        ):
            testangle = 1
            # norm of rotation angle in radians
            om = np.linalg.norm(params.slices.slice_angle[index_level])
            # unit vector marking the rotation axis
            axis_rot = params.slices.slice_angle[index_level] / om
            cos_om = np.cos(np.pi * om / 180.0)
            sin_om = np.sin(np.pi * om / 180.0)
            coeff = axis_rot[2] * (1 - cos_om)
            norm_plane = np.zeros(3)
            norm_plane[0] = axis_rot[0] * coeff + axis_rot[1] * sin_om
            norm_plane[1] = axis_rot[1] * coeff - axis_rot[0] * sin_om
            norm_plane[2] = axis_rot[2] * coeff + cos_om
            Z0 = (
                np.dot(norm_plane, (params.slices.zslice_coord[index_level]))
                / norm_plane[2]
            )
        else:
            # horizontal plane z=cte
            Z0 = params.slices.zslice_coord[index_level][2]
        Z0virt = Z0
        if hasattr(params, "interface_coord") and hasattr(
            params, "refraction_index"
        ):
            H = params.interface_coord[2]
            if H > Z0:
                # corrected z (virtual object)
                Z0virt = H - (H - Z0) / params.refraction_index
                test_refraction = 1

        if hasattr(params, "f") is False:
            params.f = np.asarray([1, 1])
        if hasattr(params, "T") is False:
            params.T = np.asarray([0, 0, 1])
        if hasattr(params, "C") is False:
            params.C = np.asarray([0, 0])
        if hasattr(params, "kc") is False:
            params.kc = 0

        if hasattr(params, "R"):
            R = copy.deepcopy(params.R)
            # R[0]= params.R[4]
            # R[1]= params.R[3]
            # R[2]= params.R[5]
            # R[4]= params.R[0]
            # R[5]= params.R[2]
            # R[6]= params.R[7]
            # R[7]= params.R[6]

            # R[1]= params.R[3]
            # R[3]= params.R[1]
            # R[2]= params.R[6]
            # R[6]= params.R[2]
            # R[5]= params.R[7]
            # R[7]= params.R[5]

            # R[1] = -R[1]
            # R[3] = -R[3]

            if testangle:
                a = -norm_plane[0] / norm_plane[2]
                b = -norm_plane[1] / norm_plane[2]
                if test_refraction:
                    a /= params.refraction_index
                    b /= params.refraction_index

                R[0] += a * R[2]
                R[1] += b * R[2]
                R[3] += a * R[5]
                R[4] += b * R[5]
                R[6] += a * R[8]
                R[7] += b * R[8]

            Tx = params.T[0]
            Ty = params.T[1]
            Tz = params.T[2]
            Dx = R[4] * R[6] - R[3] * R[7]
            Dy = R[0] * R[7] - R[1] * R[6]
            D0 = R[1] * R[3] - R[0] * R[4]
            Z11 = R[5] * R[7] - R[4] * R[8]
            Z12 = R[1] * R[8] - R[2] * R[7]
            Z21 = R[3] * R[8] - R[5] * R[6]
            Z22 = R[2] * R[6] - R[0] * R[8]
            Zx0 = R[2] * R[4] - R[1] * R[5]
            Zy0 = R[0] * R[5] - R[2] * R[3]

            A11 = R[7] * Ty - R[4] * Tz + Z11 * Z0virt
            A12 = R[1] * Tz - R[7] * Tx + Z12 * Z0virt
            A21 = -R[6] * Ty + R[3] * Tz + Z21 * Z0virt
            A22 = -R[0] * Tz + R[6] * Tx + Z22 * Z0virt

            X0 = R[4] * Tx - R[1] * Ty + Zx0 * Z0virt
            Y0 = -R[3] * Tx + R[0] * Ty + Zy0 * Z0virt

            Xd = (X - params.C[0]) / params.f[0]  # sensor coordinates
            Yd = (Y - params.C[1]) / params.f[1]
            dist_fact = 1 + params.kc * (Xd * Xd + Yd * Yd)
            Xu = Xd / dist_fact  # undistorted sensor coordinates
            Yu = Yd / dist_fact
            denom = Dx * Xu + Dy * Yu + D0
            Xphys = (A11 * Xu + A12 * Yu + X0) / denom  # world coordinates
            Yphys = (A21 * Xu + A22 * Yu + Y0) / denom
            if testangle:
                Zphys = Z0 + a * Xphys + b * Yphys
            else:
                Zphys = Z0
        else:
            Xphys = -params.T[0] + X / params.f[0]
            Yphys = -params.T[1] + Y / params.f[1]
            Zphys = X * 0
        Xphys /= 100  # cm to m
        Yphys /= 100  # cm to m
        Zphys /= 100  # cm to m

        return Xphys, Yphys, Zphys

    def phys2pix(self, Xphys, Yphys, Zphys=0):
        params = copy.deepcopy(self.params)
        Xphys *= 100  # m to cm
        Yphys *= 100  # m to cm
        Zphys *= 100  # m to cm

        if hasattr(params, "f") is False:
            params.f = np.asarray([1, 1])
        if hasattr(params, "T") is False:
            params.T = np.asarray([0, 0, 1])

        # general case
        if hasattr(params, "R"):
            R = params.R
            if hasattr(params, "interface_coord") and hasattr(
                params, "refraction_index"
            ):
                H = params.interface_coord[2]
                if H > Zphys:
                    Zphys = H - (H - Zphys) / params.refraction_index
            # corrected z (virtual object)

            # camera coordinates
            xc = R[0] * Xphys + R[1] * Yphys + R[2] * Zphys + params.T[0]
            yc = R[3] * Xphys + R[4] * Yphys + R[5] * Zphys + params.T[1]
            zc = R[6] * Xphys + R[7] * Yphys + R[8] * Zphys + params.T[2]

            # undistorted image coordinates
            Xu = xc / zc
            Yu = yc / zc

            # radial quadratic correction factor
            if hasattr(params, "kc") is False:
                r2 = 1  # no quadratic distortion
            else:
                r2 = 1 + params.kc * (Xu * Xu + Yu * Yu)

            # pixel coordinates
            if hasattr(params, "C") is False:
                params.C = np.asarray([0, 0])  # default value

            X = params.f[0] * Xu * r2 + params.C[0]
            Y = params.f[1] * Yu * r2 + params.C[1]

        # case 'rescale'
        else:
            X = params.f[0] * (Xphys + params.T[0])
            Y = params.f[1] * (Yphys + params.T[1])

        return X, Y



"""Execute a topology sequentially (:mod:`fluidimage.executors.exec_async_sequential`)
======================================================================================

.. autoclass:: ExecutorBase
   :members:
   :private-members:

"""

import time

from fluidimage.util import cstring, log_memory_usage, logger

from .base import ExecutorBase


class ExecutorSequential(ExecutorBase):
    """Execute a topology sequentially"""

    def compute(self):
        """Compute the whole topology."""
        self._init_compute()
        self.exec_one_shot_works()
        self._run_works()
        self._finalize_compute()

    def _run_works(self):

        while not all([len(queue) == 0 for queue in self.topology.queues]):

            for work in self.works:

                # global functions
                if work.kind is not None and "global" in work.kind:
                    if len(work.output_queue) > self.nb_items_queue_max:
                        continue

                    work.func_or_cls(work.input_queue, work.output_queue)

                else:
                    if not work.input_queue:
                        continue

                    key, obj = work.input_queue.pop_first_item()

                    if work.check_exception(key, obj):
                        continue

                    t_start = time.time()
                    log_memory_usage(
                        f"{time.time() - self.t_start:.2f} s. Launch work "
                        + work.name_no_space
                        + f" ({key}). mem usage"
                    )
                    # pylint: disable=W0703
                    try:
                        ret = work.func_or_cls(obj)
                    except Exception as error:
                        self.log_exception(error, work.name_no_space, key)
                        if self.stop_if_error:
                            raise
                        ret = error
                    else:
                        logger.info(
                            f"work {work.name_no_space} ({key}) "
                            f"done in {time.time() - t_start:.3f} s"
                        )

                    if work.output_queue is not None:
                        work.output_queue[key] = ret



import matplotlib.pyplot as plt
import numpy as np
from numpy import pi

from fluidimage.calcul.interpolate.thin_plate_spline2 import ThinPlateSpline


def myplot(i, x, y, U):
    plt.figure(i)
    ax = plt.gca()

    ax.scatter(x, y, c=U, vmin=-1, vmax=1)


# set of random x coordinates from 0 to 2pi
x = 2 * np.pi * np.random.rand(100)
y = 2 * np.pi * np.random.rand(100)

U = np.exp(-((x - pi) ** 2 + (y - pi) ** 2))  # gaussian
# U = x  # linear

myplot(0, x, y, U)


# calculate tps coeff
centers = np.vstack([x, y])
smoothing_coef = 0
subdom_size = 20


# interpolation grid

xI = yI = np.arange(0, 2 * pi, 0.1)
XI, YI = np.meshgrid(xI, yI)
XI = XI.ravel()
YI = YI.ravel()

new_positions = np.vstack([XI, YI])

tps = ThinPlateSpline(
    new_positions,
    centers,
    U,
    subdom_size,
    smoothing_coef,
    threshold=1,
    pourc_buffer_area=0.5,
)

U_eval = tps.compute_U_eval()
DMX_eval, DMY_eval = tps.compute_dxy_eval()


myplot(1, XI, YI, U_eval)

myplot(2, XI, YI, DMX_eval)
myplot(3, XI, YI, DMY_eval)


plt.show()



import numpy as np
import pylab

from fluiddyn.util.serieofarrays import SeriesOfArrays
from fluidimage.works.piv import WorkPIV
from postproc import PIV_Postproc, PIV_PostProc_serie

# %%
params = WorkPIV.create_default_params()

# for a very short computation
params.piv0.shape_crop_im0 = 24
params.piv0.grid.overlap = 0.0

params.piv0.method_subpix = "2d_gaussian"
params.piv0.method_correl = "fftw"

params.multipass.number = 3
# params.multipass.use_tps = 'last'

piv = WorkPIV(params=params)

series = SeriesOfArrays("../../image_samples/Karman/Images", "i+1:i+3")
serie = series.get_serie_from_index(0)

result = piv.calcul(serie)
result.save()
# result.display()

lightresult = result.make_light_result()

lightresult.save()


postp = PIV_Postproc(path="piv_Karman01-02_light.h5")

rot = postp.compute_rot()

div = postp.compute_div()

pylab.figure
postp.displayf(X=postp.X, Y=postp.Y, U=postp.U, V=postp.V)

pylab.figure
postp.displayf(bg=rot)


postp = PIV_PostProc_serie(path=["piv_Karman01-02_light.h5"] * 10)
postp.set_time(np.linspace(0, 10, 10))

rot = postp.compute_rot()

div = postp.compute_div()

postp.compute_temporal_fft()

pylab.figure
postp.displayf(U=postp.U, V=postp.V, bg=div)

pylab.figure
postp.displayf(bg=rot)

pylab.figure
postp.displayf(U=postp.U, V=postp.V)


Nt = 100
postp = PIV_PostProc_serie(path=["piv_Karman01-02_light.h5"] * Nt)

t = np.linspace(0, 1000, Nt)

postp.set_time(t)

kx = 0.05
ky = 0.17
omega = 0.02

for i in range(np.shape(postp.U)[0]):
    postp.U[i] = np.cos(omega * t[i] + kx * postp.X + ky * postp.Y)
    postp.V[i] = np.cos(omega * t[i] + kx * postp.X + ky * postp.Y)


# FFT temporelle
postp.compute_temporal_fft(parseval=True)
omega = postp.fft.time.omega
psd = postp.fft.time.psdU + postp.fft.time.psdV
pylab.loglog(omega, postp.spatial_average(psd))

# FFT spatiale
postp.compute_spatial_fft(parseval=True)
Kx, Ky = np.meshgrid(postp.fft.spatial.kx, postp.fft.spatial.ky)
Kx = Kx.transpose()
Ky = Ky.transpose()
psd = postp.fft.spatial.psdU + postp.fft.spatial.psdV
postp.displayf(X=Kx, Y=Ky, bg=np.log(postp.time_average(psd)))


# FFT spatiotemporelle
postp.compute_spatiotemp_fft(parseval=True)
Kx, Ky = np.meshgrid(postp.fft.spatiotemp.kx, postp.fft.spatiotemp.ky)
Kx = Kx.transpose()
Ky = Ky.transpose()
omega = postp.fft.spatiotemp.omega
psd = postp.fft.spatiotemp.psdU + postp.fft.spatiotemp.psdV
psd = psd[0:10:]
postp.displayf(X=Kx, Y=Ky, bg=np.log(psd))


# useful quantities
X = postp.X
Y = postp.Y
dx = X[1][0] - X[0][0]
dy = Y[0][1] - Y[0][0]
Lx = np.max(X) - np.min(X)
Ly = np.max(Y) - np.min(Y)
nx = X.shape[0]
ny = X.shape[1]

kx = postp.fft.spatiotemp.kx
ky = postp.fft.spatiotemp.ky
dkx = kx[1] - kx[0]
dky = ky[1] - ky[0]
Kx, Ky = np.meshgrid(postp.fft.spatiotemp.kx, postp.fft.spatiotemp.ky)
Kx = Kx.transpose()
Ky = Ky.transpose()

omega = postp.fft.spatiotemp.omega
domega = omega[1] - omega[0]
dt = postp.t[1] - postp.t[0]
Lt = np.max(postp.t) - np.min(postp.t)


# moyenne spatial de fft spatiotemp vs fft temp

psd = (
    np.sum(postp.fft.spatiotemp.psdU + postp.fft.spatiotemp.psdV, (1, 2))
    * 1.0
    * dkx
    * dky
)
pylab.loglog(omega, psd)
psd2 = postp.fft.time.psdU + postp.fft.time.psdV
psd2 = np.sum(psd2, (1, 2)) * 1.0 * (dx * dy) / (Lx * Ly)
pylab.loglog(omega, psd2, "r+")
print(np.mean(((psd2 - psd) ** 2) / psd))
print(np.max(((psd2 - psd) ** 2) / psd))

# moyenne temporelle de fft spatiotemp vs fft spatial
psd = (
    np.sum(postp.fft.spatiotemp.psdU + postp.fft.spatiotemp.psdV, 0)
    * 1.0
    * domega
)
pylab.pcolor(Kx, Ky, psd)
psd2 = postp.fft.spatial.psdU + postp.fft.spatial.psdV
psd2 = np.sum(psd2, 0) * 1.0 * dt / Lt
pylab.pcolor(Kx, Ky, psd2)
pylab.pcolor(Kx, Ky, psd2 - psd)
print(np.mean(((psd2 - psd) ** 2) / psd))
print(np.max(((psd2 - psd) ** 2) / psd))



import os
import sys
import unittest

from fluiddyn.io import stdout_redirected
from fluidimage import get_path_image_samples
from fluidimage.run_from_xml import main

path_image_samples = get_path_image_samples()


class TestRunFromXML(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.current_dir = os.getcwd()
        os.chdir(path_image_samples)

    @classmethod
    def tearDownClass(cls):
        os.chdir(cls.current_dir)

    def test_main(self):

        path = path_image_samples / "Karman/Images.civ/0_XML/Karman_1-4.xml"
        command = f"run {str(path)} --mode recompute"
        sys.argv = command.split()

        with stdout_redirected():
            main()



import matplotlib.pyplot as plt
import numpy as np
import scipy as sc
from matplotlib.mlab import griddata
from numpy import pi
from scipy import interpolate

from fluidimage.calcul.interpolate.thin_plate_spline import (
    ThinPlateSpline,
    compute_tps_coeff,
)

plt.ion()


n0 = 100
n1 = 100

# method = 'griddata_scipy'
method = "tps"


def myplot(i, x, y, U, title=None):
    plt.figure(i)
    ax = plt.gca()

    ax.scatter(x, y, c=U, vmin=-1, vmax=1)

    if title is not None:
        ax.set_title(title)


# set of random x coordinates from 0 to 2pi
x = 2 * np.pi * np.random.rand(n0)
y = 2 * np.pi * np.random.rand(n0)

U = np.exp(-((x - pi) ** 2 + (y - pi) ** 2))  # gaussian
# U = x  # linear

myplot(0, x, y, U, "input data")

centers = np.vstack([x, y])
# interpolation grid
xI = yI = np.linspace(0, 2 * pi, n1)
XI, YI = np.meshgrid(xI, yI)
XI = XI.ravel()
YI = YI.ravel()

new_positions = np.vstack([XI, YI])

if method == "tps":

    # calculate tps coeff
    U_smooth, U_tps = compute_tps_coeff(centers, U, 0)
    # evaluate interpolation on the new grid
    tps = ThinPlateSpline(new_positions, centers)

    U_eval = tps.compute_field(U_tps)

    # U_eval = griddata(x,y,U,xI,yI,'linear') #griddata method from matplotlib

    DUX_eval, DUY_eval = tps.compute_gradient(U_tps)

elif method == "griddata_mpl":

    U_eval = griddata(x, y, U, xI, yI, "linear")

    DUX_eval, DUY_eval = sc.gradient(U_eval)

elif method == "griddata_scipy":

    grid_x, grid_y = np.meshgrid(xI, yI)
    # nearest, linear, cubic
    U_eval = interpolate.griddata(centers.T, U, (grid_x, grid_y), "cubic", 0)

    DUX_eval, DUY_eval = sc.gradient(U_eval)


myplot(1, XI, YI, U_eval, "U_eval")

myplot(2, XI, YI, DUX_eval, "DUX_eval")
myplot(3, XI, YI, DUY_eval, "DUY_eval")

plt.show()



import os
import unittest
from glob import glob
from shutil import rmtree

from fluiddyn.io import stdout_redirected
from fluiddyn.io.image import imread, imsave
from fluidimage import get_path_image_samples
from fluidimage.preproc.base import PreprocBase


class TestPreprocKarman(unittest.TestCase):

    name = "Karman"

    @classmethod
    def setUpClass(cls):
        path_in = str(get_path_image_samples() / cls.name / "Images")

        cls._work_dir = os.path.join(
            "test_fluidimage_topo_preproc_" + cls.name, "Images"
        )
        if not os.path.exists(cls._work_dir):
            os.makedirs(cls._work_dir)

        paths = glob(path_in + "/*")

        for path in paths:
            name = os.path.split(path)[-1]
            im = imread(path)
            im = im[::6, ::6]
            imsave(os.path.join(cls._work_dir, name), im, as_int=True)

    @classmethod
    def tearDownClass(cls):
        rmtree(os.path.split(cls._work_dir)[0], ignore_errors=True)

    def test_preproc(self):
        """Test preproc subpackage on image sample Karman with one index."""
        params = PreprocBase.create_default_params()

        params.preproc.series.path = self._work_dir

        for tool in params.preproc.tools.available_tools:
            if "sliding" not in tool and "temporal" not in tool:
                tool = params.preproc.tools.__getitem__(tool)
                tool.enable = True

        preproc = PreprocBase(params)
        preproc()
        preproc.display(1, hist=True)


class TestPreprocTime(TestPreprocKarman):
    name = "Jet"

    def test_preproc(self):
        """Test preproc subpackage on image sample Jet with two indices."""
        params = PreprocBase.create_default_params()

        params.preproc.series.path = self._work_dir

        for tool in params.preproc.tools.available_tools:
            if "sliding" in tool:
                tool = params.preproc.tools.__getitem__(tool)
                tool.enable = True

        preproc = PreprocBase(params)
        preproc()
        preproc.display(1, hist=False)


if __name__ == "__main__":
    unittest.main()



import ctypes
import time

import numpy as np

import pycuda.autoinit
import pycuda.gpuarray as gpuarray

CUSOLVER_STATUS_SUCCESS = 0

libcusolver = ctypes.cdll.LoadLibrary("libcusolver.so")

libcusolver.cusolverDnCreate.restype = int
libcusolver.cusolverDnCreate.argtypes = [ctypes.c_void_p]


def cusolverDnCreate():
    handle = ctypes.c_void_p()
    status = libcusolver.cusolverDnCreate(ctypes.byref(handle))
    if status != CUSOLVER_STATUS_SUCCESS:
        raise RuntimeError("error!")

    return handle.value


libcusolver.cusolverDnDestroy.restype = int
libcusolver.cusolverDnDestroy.argtypes = [ctypes.c_void_p]


def cusolverDnDestroy(handle):
    status = libcusolver.cusolverDnDestroy(handle)
    if status != CUSOLVER_STATUS_SUCCESS:
        raise RuntimeError("error!")


libcusolver.cusolverDnSgetrf_bufferSize.restype = int
libcusolver.cusolverDnSgetrf_bufferSize.argtypes = [
    ctypes.c_void_p,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_void_p,
    ctypes.c_int,
    ctypes.c_void_p,
]


def cusolverDnSgetrf_bufferSize(handle, m, n, A, lda, Lwork):
    status = libcusolver.cusolverDnSgetrf_bufferSize(
        handle, m, n, int(A.gpudata), n, ctypes.pointer(Lwork)
    )
    if status != CUSOLVER_STATUS_SUCCESS:
        raise RuntimeError("error!")


libcusolver.cusolverDnSgetrf.restype = int
libcusolver.cusolverDnSgetrf.argtypes = [
    ctypes.c_void_p,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_void_p,
    ctypes.c_int,
    ctypes.c_void_p,
    ctypes.c_void_p,
    ctypes.c_void_p,
]


def cusolverDnSgetrf(handle, m, n, A, lda, Workspace, devIpiv, devInfo):
    status = libcusolver.cusolverDnSgetrf(
        handle,
        m,
        n,
        int(A.gpudata),
        lda,
        int(Workspace.gpudata),
        int(devIpiv.gpudata),
        int(devInfo.gpudata),
    )
    if status != CUSOLVER_STATUS_SUCCESS:
        raise RuntimeError("error!")


libcusolver.cusolverDnSgetrs.restype = int
libcusolver.cusolverDnSgetrs.argtypes = [
    ctypes.c_void_p,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_int,
    ctypes.c_void_p,
    ctypes.c_int,
    ctypes.c_void_p,
    ctypes.c_void_p,
    ctypes.c_int,
    ctypes.c_void_p,
]


def cusolverDnSgetrs(handle, trans, n, nrhs, A, lda, devIpiv, B, ldb, devInfo):
    status = libcusolver.cusolverDnSgetrs(
        handle,
        trans,
        n,
        nrhs,
        int(A.gpudata),
        lda,
        int(devIpiv.gpudata),
        int(B.gpudata),
        ldb,
        int(devInfo.gpudata),
    )
    if status != CUSOLVER_STATUS_SUCCESS:
        raise RuntimeError("error!")


if __name__ == "__main__":
    import numpy as np

    m = 6400
    n = 6400
    a = np.asarray(np.random.rand(m, n), np.float32)
    init_time = time.time()
    a_gpu = gpuarray.to_gpu(a.T.copy())
    lda = m
    b = np.asarray(np.random.rand(m, n), np.float32)
    b_gpu = gpuarray.to_gpu(b.T.copy())
    ldb = m

    handle = cusolverDnCreate()
    Lwork = ctypes.c_int()

    cusolverDnSgetrf_bufferSize(handle, m, n, a_gpu, lda, Lwork)
    Workspace = gpuarray.empty(Lwork.value, dtype=np.float32)
    devIpiv = gpuarray.zeros(min(m, n), dtype=np.int32)
    devInfo = gpuarray.zeros(1, dtype=np.int32)

    cusolverDnSgetrf(handle, m, n, a_gpu, lda, Workspace, devIpiv, devInfo)
    if devInfo.get()[0] != 0:
        raise RuntimeError("error!")

    CUBLAS_OP_N = 0
    nrhs = n
    devInfo = gpuarray.zeros(1, dtype=np.int32)
    cusolverDnSgetrs(
        handle, CUBLAS_OP_N, n, nrhs, a_gpu, lda, devIpiv, b_gpu, ldb, devInfo
    )

    x_cusolver = b_gpu.get().T
    cusolverDnDestroy(handle)
    cusolve_time = time.time() - init_time
    print("cusolve time = %.6f" % cusolve_time)
    x_numpy = np.linalg.solve(a, b)
    numpy_time = time.time() - init_time - cusolve_time
    speedup = numpy_time / cusolve_time
    print("np.linalg.solve time = %.6f" % numpy_time)
    print("GPU speedup = %f" % speedup)
    print(np.allclose(x_numpy, x_cusolver, rtol=1e-02, atol=1e-04))



"""Topology example (:mod:`fluidimage.topologies.example`)
==========================================================

This topology has two pythran cpu bounded tasks. It helps see executors behavior
with C functions.

.. autoclass:: TopologyExample
   :members:
   :private-members:

"""

import os

import numpy as np
import scipy.io

from transonic import boost

from . import TopologyBase
from ..util import imread

A = "uint8[:,:]"


@boost
def cpu1(array1: A, array2: A, nloops: int = 10):

    a = np.arange(10000000 // nloops)
    result = a
    for i in range(nloops):
        result += a**3 + a**2 + 2

    for i in range(nloops):
        array1 = array1 * array2
    return (array1, array1)


@boost
def cpu2(array1: A, array2: A, nloops: int = 10):

    a = np.arange(10000000 // nloops)
    result = a
    for i in range(nloops):
        result += a**3 + a**2 + 2

    for i in range(nloops):
        array1 = np.multiply(array1, array2)
    return array1


class TopologyExample(TopologyBase):
    """Topology example for testing.

    Parameters
    ----------

    logging_level : str, {'warning', 'info', 'debug', ...}

      Logging level.

    nb_max_workers : None, int

      Maximum numbers of "workers". If None, a number is computed from the
      number of cores detected. If there are memory errors, you can try to
      decrease the number of workers.

    """

    @classmethod
    def create_default_params(cls):
        params = dict(
            path_input=None,
            path_dir_result=None,
            nloops=1,
            multiplicator_nb_images=1,
        )
        return params

    def __init__(self, params, logging_level="info", nb_max_workers=None):
        self.params = params

        path_input = params["path_input"]
        path_dir_result = params["path_dir_result"]
        nloops = params["nloops"]
        self.multiplicator_nb_images = params["multiplicator_nb_images"]

        def func1(arrays):
            key = arrays[0]
            if key == "Karman_02_00":
                raise ValueError("For testing")
            arr0, arr1 = cpu1(arrays[1], arrays[2], nloops)
            return key, arr0, arr1

        def func2(arrays):
            key = arrays[0]
            result = cpu2(arrays[1], arrays[2], nloops)
            return key, result

        self.path_input = path_input

        super().__init__(
            path_dir_result=path_dir_result,
            logging_level=logging_level,
            nb_max_workers=nb_max_workers,
        )

        if not self.path_dir_result.exists():
            self.path_dir_result.mkdir()

        self.img_counter = 0

        queue_names = self.add_queue("names images")
        queue_couples_names = self.add_queue("names couples")
        queue_arrays = self.add_queue("arrays")
        queue_couples_arrays = self.add_queue("couples arrays")
        queue_cpu1 = self.add_queue("queue cpu1")
        queue_cpu2 = self.add_queue("queue cpu2")

        self.add_work(
            "fill names",
            func_or_cls=self.fill_names,
            output_queue=queue_names,
            kind=("global", "one shot"),
        )
        self.add_work(
            "fill names couples",
            func_or_cls=self.fill_couples_names,
            input_queue=queue_names,
            output_queue=queue_couples_names,
            kind=("global", "one shot"),
        )

        self.add_work(
            "read array",
            func_or_cls=self.read_array,
            input_queue=queue_names,
            output_queue=queue_arrays,
            kind="io",
        )
        self.add_work(
            "fill couples arrays",
            func_or_cls=self.fill_couples_arrays,
            input_queue=(queue_couples_names, queue_arrays),
            output_queue=queue_couples_arrays,
            kind=("global"),
        )
        self.add_work(
            "cpu1",
            func_or_cls=func1,
            input_queue=queue_couples_arrays,
            output_queue=queue_cpu1,
            kind="cpu",
        )
        self.add_work(
            "cpu2",
            func_or_cls=func2,
            input_queue=queue_cpu1,
            output_queue=queue_cpu2,
            kind="cpu",
        )
        self.add_work(
            "save", func_or_cls=self.save, input_queue=queue_cpu2, kind="io"
        )

    def fill_names(self, input_queue, output_queue):
        for ind in range(self.multiplicator_nb_images):
            for name in sorted(os.listdir(self.path_input)):
                key = name.split(".bmp")[0] + f"_{ind:02}"
                output_queue[key] = name

    def fill_couples_names(self, input_queue, output_queue):
        for key, name in list(input_queue.items()):
            output_queue[key] = [key, (name, name)]

    def read_array(self, name):
        if name == "Karman_03.bmp":
            raise ValueError("For testing")

        image = imread(self.path_input / name)
        return image

    def fill_couples_arrays(self, input_queues, output_queue):
        queue_couples_names, queue_arrays = input_queues
        queue_couples_arrays = output_queue

        for key, array in list(queue_arrays.items()):
            if key not in queue_couples_names:
                continue
            queue_arrays.pop(key)
            queue_couples_names.pop(key)
            # propagating possible exceptions...
            if isinstance(array, Exception):
                queue_couples_arrays[key] = array
            else:
                queue_couples_arrays[key] = [key, array, array]

    def save(self, inputs):
        key = inputs[0]
        arr = inputs[1]
        name = key + ".h5"
        scipy.io.savemat(self.path_dir_result / name, mdict={"array": arr})



"""Particle Image Velocimetry (PIV) (:mod:`fluidimage.works.piv`)
=================================================================

.. autosummary::
   :toctree:

   multipass
   singlepass
   fix

"""

from .multipass import WorkPIV

__all__ = ["WorkPIV"]



import unittest
from pathlib import Path
from shutil import rmtree

from fluidimage import get_path_image_samples
from fluidimage.topologies.image2image import TopologyImage2Image


class TestPivNew(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.path_src = get_path_image_samples() / "Karman/Images"
        cls.postfix = "test_im2im_new"

    @classmethod
    def tearDownClass(cls):
        path_out = Path(str(cls.path_src) + "." + cls.postfix)
        if path_out.exists():
            rmtree(path_out)

    def test_piv_new(self):
        params = TopologyImage2Image.create_default_params()

        params.images.path = str(self.path_src)

        params.im2im = "fluidimage.preproc.image2image.Im2ImExample"
        params.args_init = ((1024, 2048), "clip")

        params.saving.how = "recompute"
        params.saving.postfix = self.postfix

        topology = TopologyImage2Image(params, logging_level="info")
        topology.compute("exec_async", stop_if_error=True)

        topology = TopologyImage2Image(params, logging_level="info")
        topology.compute()

        topology.make_code_graphviz(topology.path_dir_result / "topo.dot")


if __name__ == "__main__":
    unittest.main()



"""Base class for executors (:mod:`fluidimage.executors.base`)
==============================================================

.. autoclass:: ExecutorBase
   :members:
   :private-members:

"""

import os
import signal
import sys
import traceback
from pathlib import Path
from time import time

from fluiddyn import time_as_str
from fluiddyn.io.tee import MultiFile
from fluidimage.config import get_config
from fluidimage.topologies.nb_cpu_cores import nb_cores
from fluidimage.util import (
    log_error,
    log_memory_usage,
    logger,
    reset_logger,
    str_short,
)

config = get_config()


class ExecutorBase:
    """Base class for executors.

    Parameters
    ----------

    topology : fluidimage.topology.base.TopologyBase

      A computational topology.

    path_dir_result : str or pathlib.Path

      The path of the directory where the results have to be saved.

    nb_max_workers : int, optional (None)

    nb_items_queue_max : int, optional (None),

    logging_level : str, optional {"info"},

    sleep_time : number, optional {None},

    stop_if_error : bool, optional {False}

    """

    def _init_log_path(self):
        name = "_".join(("log", time_as_str(), str(os.getpid())))
        self.path_dir_exceptions = self.path_dir_result / name
        self._log_path = self.path_dir_result / (name + ".txt")

    def __init__(
        self,
        topology,
        path_dir_result,
        nb_max_workers=None,
        nb_items_queue_max=None,
        logging_level="info",
        sleep_time=None,
        stop_if_error=False,
    ):
        del sleep_time
        self.topology = topology
        self.logging_level = logging_level
        self.stop_if_error = stop_if_error

        path_dir_result = Path(path_dir_result)
        path_dir_result.mkdir(exist_ok=True)
        self.path_dir_result = path_dir_result
        self._init_log_path()
        self._log_file = open(self._log_path, "w")

        stdout = sys.stdout
        if isinstance(stdout, MultiFile):
            stdout = sys.__stdout__

        stderr = sys.stderr
        if isinstance(stderr, MultiFile):
            stderr = sys.__stderr__

        sys.stdout = MultiFile([stdout, self._log_file])
        sys.stderr = MultiFile([stderr, self._log_file])

        if logging_level:
            for handler in logger.handlers:
                logger.removeHandler(handler)

            from fluidimage import config_logging

            config_logging(logging_level, file=sys.stdout)

        if nb_max_workers is None:
            if config is not None:
                try:
                    nb_max_workers = eval(config["topology"]["nb_max_workers"])
                except KeyError:
                    pass

        # default nb_max_workers
        # Difficult: trade off between overloading and limitation due to input
        # output.  The user can do much better for a specific case.
        if nb_max_workers is None:
            if nb_cores < 16:
                nb_max_workers = nb_cores + 2
            else:
                nb_max_workers = nb_cores

        self.nb_max_workers = nb_max_workers

        if nb_items_queue_max is None:
            nb_items_queue_max = max(4 * nb_max_workers, 8)
        self.nb_items_queue_max = nb_items_queue_max

        self._has_to_stop = False
        if sys.platform != "win32":

            def handler_signals(signal_number, stack):
                del stack
                print(
                    f"signal {signal_number} received: set _has_to_stop to True "
                    f"({type(self).__name__})."
                )
                self._has_to_stop = True

            signal.signal(12, handler_signals)

        # Picks up async works
        self.works = [
            work
            for work in self.topology.works
            if work.kind is None or "one shot" not in work.kind
        ]

        # to avoid a pylint warning
        self.t_start = None

    def _init_compute(self):
        self.t_start = time()
        self._init_compute_log()

    def _init_compute_log(self):
        log_memory_usage(time_as_str(2) + ": starting execution. mem usage")
        logger.info(f"  topology: {str_short(type(self.topology))}")
        logger.info(f"  executor: {str_short(type(self))}")
        logger.info(f"  nb_cpus_allowed = {nb_cores}")
        logger.info(f"  nb_max_workers = {self.nb_max_workers}")
        logger.info(f"  path_dir_result = {self.path_dir_result}")

    def _reset_std_as_default(self):
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__
        reset_logger()
        self._log_file.close()

    def _finalize_compute(self):
        log_memory_usage(time_as_str(2) + ": end of `compute`. mem usage")
        self.topology.print_at_exit(time() - self.t_start)
        self._reset_std_as_default()

    def exec_one_shot_works(self):
        """
        Execute all "one shot" functions.

        """
        for work in self.topology.works:
            if work.kind is not None and "one shot" in work.kind:
                pretty = str_short(work.func_or_cls.__func__)
                logger.info(f'Running "one_shot" job "{work.name}" ({pretty})')
                work.func_or_cls(work.input_queue, work.output_queue)

    def log_exception(self, exception, work_name, key):
        """Log an exception in a file."""
        path_log = self.path_dir_exceptions / f"exception_{work_name}_{key}.txt"

        log_error(
            "error during work " f"{work_name} ({key}) (logged in {path_log})"
        )

        self.path_dir_exceptions.mkdir(exist_ok=True)

        try:
            parts = traceback.format_exception(
                etype=type(exception), value=exception, tb=exception.__traceback__
            )
        except TypeError:
            # Python >= 3.10
            parts = traceback.format_exception(exception)
        formated_exception = "".join(parts)

        with open(path_log, "w") as file:
            file.write(
                f"Exception for work {work_name}, key {key}:\n"
                + formated_exception
            )



"""Fix work
===========

.. autoclass:: WorkFIX
   :members:
   :private-members:

"""

import numpy as np

from fluiddyn.util.paramcontainer import ParamContainer

from .. import BaseWork
from ...calcul.smooth_clean import smooth_clean


class WorkFIX(BaseWork):
    """Fix a displacement vector field."""

    @classmethod
    def create_default_params(cls):
        params = ParamContainer(tag="params")
        cls._complete_params_with_default(params)
        return params

    @classmethod
    def _complete_params_with_default(cls, params, tag="fix"):

        params._set_child(
            tag,
            attribs={
                "correl_min": 0.2,
                "threshold_diff_neighbour": 10,
                "displacement_max": None,
            },
        )

        params.fix._set_doc(
            """
Parameters indicating how are detected and processed false vectors.

correl_min : 0.2

    Vectors associated with correlation smaller than correl_min are considered
    as false vectors.

threshold_diff_neighbour : 10

    Vectors for which the difference with the average vectors is larger than
    `threshold_diff_neighbour` are considered as false vectors.

displacement_max : None

    Vectors larger than `displacement_max` are considered as false vectors.

"""
        )

    def __init__(self, params, piv_work):
        self.params = params
        self.piv_work = piv_work

    def calcul(self, piv_results):
        # print('piv_results = ', piv_results)
        deltaxs_wrong = {}
        deltays_wrong = {}

        deltaxs = piv_results.deltaxs
        deltays = piv_results.deltays
        for ierr in piv_results.errors.keys():
            deltaxs_wrong[ierr] = deltaxs[ierr]
            deltays_wrong[ierr] = deltays[ierr]
            deltaxs[ierr] = np.nan
            deltays[ierr] = np.nan

        def put_to_nan(inds, explanation):
            for ind in inds:
                ind = int(ind)

                if ind not in deltaxs_wrong:
                    deltaxs_wrong[ind] = deltaxs[ind]
                    deltays_wrong[ind] = deltays[ind]

                    deltaxs[ind] = np.nan
                    deltays[ind] = np.nan
                    piv_results.errors[ind] = explanation
                else:
                    piv_results.errors[ind] += " + " + explanation

        # condition correl < correl_min
        inds = (piv_results.correls_max < self.params.correl_min).nonzero()[0]
        put_to_nan(inds, "correl < correl_min")

        # condition delta2 < displacement_max2
        if self.params.displacement_max:
            displacement_max2 = self.params.displacement_max**2
            delta2s = deltaxs**2 + deltays**2
            with np.errstate(invalid="ignore"):
                inds = (delta2s > displacement_max2).nonzero()[0]
            put_to_nan(inds, "delta2 < displacement_max2")

        # warning condition neighbour not implemented...
        if self.params.threshold_diff_neighbour is not None:
            threshold = self.params.threshold_diff_neighbour
            ixvecs = self.piv_work.ixvecs
            iyvecs = self.piv_work.iyvecs
            xs = piv_results.xs
            ys = piv_results.ys

            # import ipdb; ipdb.set_trace()
            ixvecs, iyvecs = self.piv_work._xyoriginalimage_from_xymasked(
                ixvecs, iyvecs
            )

            dxs_smooth, dys_smooth = smooth_clean(
                xs, ys, deltaxs, deltays, iyvecs, ixvecs, threshold
            )
            piv_results.dxs_smooth_clean = dxs_smooth
            piv_results.dys_smooth_clean = dys_smooth

            differences = np.sqrt(
                (dxs_smooth - deltaxs) ** 2 + (dys_smooth - deltays) ** 2
            )

            with np.errstate(invalid="ignore"):
                inds = (differences > threshold).nonzero()[0]

            put_to_nan(inds, "diff neighbour too large")

            for ind in inds:
                piv_results.errors[ind] += " (diff = {:.2f})".format(
                    differences[ind]
                )

            piv_results.deltaxs_wrong = deltaxs_wrong
            piv_results.deltays_wrong = deltays_wrong

        # condition check 2d peak
        ratio_correl_peaks = 0.85

        secondary_peaks = piv_results.secondary_peaks
        for ind, other_peaks in enumerate(secondary_peaks):
            if other_peaks is None or len(other_peaks) == 0:
                continue

            correl0 = piv_results.correls_max[ind]

            other_peaks_good = []
            for (dx, dy, corr) in other_peaks:
                if (
                    corr / correl0 > ratio_correl_peaks
                    and corr > self.params.correl_min
                ):
                    other_peaks_good.append((dx, dy, corr))

            if len(other_peaks_good) == 0:
                continue

            diff_neighbours = np.empty(len(other_peaks_good) + 1)
            diff_neighbours[0] = differences[ind]

            for i, (dx, dy, corr) in enumerate(other_peaks_good):
                diff_neighbours[i + 1] = np.sqrt(
                    (dxs_smooth[ind] - dx) ** 2 + (dys_smooth[ind] - dy) ** 2
                )

            argmin = diff_neighbours.argmin()

            if argmin != 0:
                print("replace peak")
                dx, dy, corr = other_peaks_good[argmin - 1]
                other_peaks_good[argmin - 1] = (
                    piv_results.deltaxs[ind],
                    piv_results.deltays[ind],
                    piv_results.correls_max[ind],
                )
                piv_results.secondary_peaks[ind] = other_peaks_good

                piv_results.deltaxs[ind] = dx
                piv_results.deltays[ind] = dy
                piv_results.correls_max[ind] = corr

                if ind in piv_results.errors:
                    del (
                        piv_results.deltaxs_wrong[ind],
                        piv_results.deltays_wrong[ind],
                    )
                    print("!!!!! TO DO ind in piv_results.errors !!!!!")

        return piv_results
